from __future__ import annotations

__doc__ = """

Data Manager (OpenEphys)
########################

.. autoclass:: DataManager
   :members:

Module (OpenEphys)
##################

.. Note::
    We expect the data structure to follow the default format
    exported from OpenEphys system:
    `format <https://open-ephys.atlassian.net/wiki/spaces/OEW/pages/491632/Data+format>`_.

.. currentmodule:: miv.io.openephys.data

.. autoclass:: Data
   :members:

"""

__all__ = ["Data", "DataManager"]


import logging
import os
import math
import re
from collections.abc import MutableSequence
from glob import glob
from typing import (
    TYPE_CHECKING,
    Any,
)
from collections.abc import Iterable, Generator

import numpy as np

from miv.core.datatype.signal import Signal
from miv.core.operator.operator import DataLoaderMixin
from miv.io.openephys.binary import load_recording, load_ttl_event
from miv.io.protocol import DataProtocol
from .binary import load_timestamps, oebin_read

if TYPE_CHECKING:
    import mpi4py


class Data(DataLoaderMixin):
    """Single data unit handler.

    Each data unit that contains single recording. This class provides useful tools,
    such as masking channel, export data, interface with other packages, etc.
    If you have multiple recordings you would like to handle at the same time, use
    `DataManager` instead.

    By default recording setup, the following directory structure is expected in ``data_path``::

        recording1                              # <- recording data_path
        ├── continuous
        │   └── Rhythm_FPGA-100.0
        │       ├── continuous.dat
        │       ├── synchronized_timestamps.npy
        │       └── timestamps.npy
        ├── events
        │   ├── Message_Center-904.0
        │   │   └── TEXT_group_1
        │   │       ├── channels.npy
        │   │       ├── text.npy
        │   │       └── timestamps.npy
        │   └── Rhythm_FPGA-100.0
        │       └── TTL_1
        │           ├── channel_states.npy
        │           ├── channels.npy
        │           ├── full_words.npy
        │           └── timestamps.npy
        ├── sync_messages.txt
        ├── structure.oebin
        └── analysis                            # <- post-processing result
            ├── spike_data.npz
            ├── plot
            ├── spike
            └── mea_overlay


    Parameters
    ----------
        data_path : str
    """

    def __init__(
        self,
        data_path: str,
        tag: str = "data",
    ) -> None:
        self.data_path: str = data_path
        self.tag: str = f"{tag}"
        super().__init__()
        self.masking_channel_set: set[int] = set()

    def num_fragments(self) -> int:
        # Refactor
        file_path: list[str] = glob(
            os.path.join(self.data_path, "**", "continuous.dat"), recursive=True
        )
        assert len(file_path) == 1, (
            f"There should be only one 'continuous.dat' file. (There exists {file_path})"
        )

        # load structure information dictionary
        info_file: str = os.path.join(self.data_path, "structure.oebin")
        info: dict[str, Any] = oebin_read(info_file)
        num_channels: int = info["continuous"][0]["num_channels"]
        sampling_rate: int = int(info["continuous"][0]["sample_rate"])
        # channel_info: dict[str, Any] = info["continuous"][0]["channels"]

        _old_oe_version = False

        # Read timestamps first
        dirname = os.path.dirname(file_path[0])
        timestamps_path = os.path.join(dirname, "timestamps.npy")
        timestamps = load_timestamps(timestamps_path, sampling_rate, _old_oe_version)
        total_length = timestamps.size

        # Define task
        filesize = os.path.getsize(file_path[0])
        itemsize = np.dtype("int16").itemsize
        assert filesize == itemsize * total_length * num_channels, (
            f"{filesize=} does not match the expected {itemsize*total_length*num_channels=}. Path: {file_path[0]}"
        )
        samples_per_block = sampling_rate * 60
        num_fragments = int(math.ceil(total_length / samples_per_block))
        return num_fragments

    @property
    def number_of_channels(self):
        info_file = os.path.join(self.data_path, "structure.oebin")
        info = oebin_read(info_file)
        return info["continuous"][0]["num_channels"]

    def load(
        self,
        start_at_zero: bool = False,
        progress_bar: bool = False,
        mpi_comm: mpi4py.MPI.Comm | None = None,
    ) -> Generator[Signal]:
        """
        Iterator to load data fragmentally.

        Parameters
        ----------
        start_at_zero : bool
            If set to True, time first timestamps will be shifted to zero. To achieve synchronized
            timestamps with other recordings/events, set this to False.
        progress_bar : bool
            Visible progress bar
        mpi_comm: Optional[MPI.COMM_WORLD]
            (Experimental Feature) If the load is executed in MPI environment, user can pass
            MPI.COMM_WORLD to split the load.
            For example, if num_fragments=100, rank=2, and size=10, this function will only iterate
            the fragment 20-29 out of 0-100.

        Examples
        --------
            >>> data = Data(data_path)
            >>> for data.load(num_fragments=10) as (signal, timestamps, sampling_rate):
            ...     ...

        Returns
        -------
        signal : SignalType, neo.core.AnalogSignal
            The length of the first axis `signal.shape[0]` correspond to the length of the
            signal, while second axis `signal.shape[1]` correspond to the number of channels.
        timestamps : numpy array
        sampling_rate : float

        Raises
        ------
        FileNotFoundError
            If some key files are missing.
        """
        # TODO: Not sure this is safe implementation
        if not self.check_path_validity():
            raise FileNotFoundError("Data directory does not have all necessary files.")

        for signal, timestamps, rate in load_recording(
            self.data_path,
            self.masking_channel_set,
            start_at_zero=start_at_zero,
            progress_bar=progress_bar,
            mpi_comm=mpi_comm,
        ):
            yield Signal(data=signal, timestamps=timestamps, rate=rate)

    def load_ttl_event(self) -> Signal:
        """
        Load TTL event data if data contains. Detail implementation is :func:`here <miv.io.binary.load_ttl_event>`.
        """
        states, full_words, timestamps, sampling_rate, initial_state = load_ttl_event(
            self.data_path
        )
        if timestamps.size == 0:
            self.logger.warning(
                f"TTL event was loaded but data is empty: {self.data_path}"
            )
        return Signal(data=states[:, None], timestamps=timestamps, rate=sampling_rate)

    def set_channel_mask(self, channel_id: Iterable[int]) -> None:
        """
        Set the channel masking.

        Parameters
        ----------
        channel_id : Iterable[int], list
            List of channel id that will be ignored.

        Notes
        -----
        If the index exceed the number of channels, it will be ignored.

        Examples
        --------
        >>> data = Data(data_path)
        >>> data.set_channel_mask(range(12,23))

        """
        self.masking_channel_set.update(channel_id)

    def clear_channel_mask(self) -> None:
        """
        Clears all present channel masks.
        """
        self.masking_channel_set = set()

    def check_path_validity(self) -> bool:
        """
        Check if necessary files exist in the directory.

        - Check `continious.dat` exists. (only one)
        - Check `structure.oebin` exists.

        Returns
        -------
        bool
            Return true if all necessary files exist in the directory.

        """
        continuous_dat_paths = glob(
            os.path.join(self.data_path, "**", "continuous.dat"), recursive=True
        )
        if len(continuous_dat_paths) != 1:
            logging.warning(
                f"One and only one continuous.dat file can exist in the data path. Found: {continuous_dat_paths}"
            )
            return False
        if not os.path.exists(os.path.join(self.data_path, "structure.oebin")):
            logging.warning("Missing structure.oebin in the data path.")
            return False
        return True


class DataManager(MutableSequence):
    """
    Data collection manager.

    By default recording setup, the directory is named after the date and time
    of the recording. The structure of ``data_collection_path`` typically look
    like below::

        2022-03-10_16-19-09         <- data_collection_path
        └── Record Node 104
            └── experiment1
                └── recording1      <- data_path (Data module)
            ├── experiment2
            ├── experiment3
            ├── experiment4
            ├── spontaneous
            ├── settings.xml
            ├── settings_2.xml
            └── settings_3.xml

    Parameters
    ----------
        data_collection_path : str
            Path for data collection.

    """

    def __init__(self, data_collection_path: str) -> None:
        self.tag = "data_collection"
        super().__init__()
        self.data_collection_path: str = data_collection_path

        # From the path get data paths and create data objects
        self.data_list: list[DataProtocol] = []
        self._load_data_paths()

    @property
    def data_path_list(self) -> Iterable[str]:
        return [data.data_path for data in self.data_list]

    # Queries
    def query_path_name(self, query_path: str) -> Iterable[DataProtocol]:
        return list(filter(lambda d: query_path in d.data_path, self.data_list))

    # DataManager Representation
    def tree(self) -> None:  # pragma: no cover
        """
        Pretty-print available recordings in DataManager in tree format.

        Examples
        --------
        >>> data_collection = DataManager("2022-05-15_13-51-36")
        >>> data_collection.tree()
        2022-05-15_14-51-36
            0: <miv.io.data.Data object at 0x7f8960660cd0>
               └── Record Node 103/experiment3_std2_pt_ESC/recording1
            1: <miv.io.data.Data object at 0x7f89671c8400>
               └── Record Node 103/experiment2_std1_pt_ESC/recording1
            2: <miv.io.data.Data object at 0x7f896199e7c0>
               └── Record Node 103/experiment1_cont_ESC/recording1

        """
        # TODO: Either use logging or other str stream
        if not self.data_list:
            print(
                "Data list is empty. Check if data_collection_path exists and correct"
            )
            return
        print(self.data_collection_path)
        for idx, data in enumerate(self.data_list):
            print(" " * 4 + f"{idx}: {data}")
            print(" " * 4 + "   └── " + data.data_path)

    def _load_data_paths(self) -> None:
        """
        Create data objects from the data three.
        """
        # From the path get the data path list
        data_path_list = self._get_experiment_paths()

        # Create data object
        self.data_list = []
        invalid_count = 0
        counter = 0
        for counter, path in enumerate(data_path_list):
            data = Data(path)
            if data.check_path_validity():
                self.data_list.append(data)
            else:
                invalid_count += 1
        logging.info(
            f"Total {counter} recording found. There are {invalid_count} invalid paths."
        )

    def _get_experiment_paths(self, sort: bool = True) -> Iterable[str]:
        """
        Get experiment paths.

        Parameters
        ----------
        sort : bool
            Sort the output data paths

        Returns
        -------
        data_path_list : list
        """
        # Use self.data_collection_path
        path_list = []
        for path in glob(
            os.path.join(self.data_collection_path, "*", "experiment*", "recording*")
        ):
            if (
                ("Record Node" in path)
                and ("experiment" in path)
                and os.path.isdir(path)
            ):
                path_list.append(path)
        if sort:
            if os.name == "nt":  # For Windows
                pattern = r"(\d+)\\recording(\d+)"
            else:  # For Linux or other POSIX systems
                pattern = r"(\d+)/recording(\d+)"
            matches = [re.search(pattern, path) for path in path_list]
            tags: list[tuple[int, int]] = []
            for match in matches:
                if match is not None:
                    tags.append((int(match.group(1)), int(match.group(2))))
            path_list = [path for _, path in sorted(zip(tags, path_list, strict=False))]
        return path_list

    # MutableSequence abstract methods
    def __len__(self):  # type: ignore[no-untyped-def]
        return len(self.data_list)

    def __getitem__(self, idx):  # type: ignore[no-untyped-def]
        return self.data_list[idx]

    def __delitem__(self, idx):  # type: ignore[no-untyped-def]
        del self.data_list[idx]

    def __setitem__(self, idx, data):  # type: ignore[no-untyped-def]
        if data.check_path_validity():
            self.data_list[idx] = data
        else:
            logging.warning("Invalid data cannot be loaded to the DataManager.")

    def insert(self, idx, data):  # type: ignore[no-untyped-def]
        if data.check_path_validity():
            self.data_list.insert(idx, data)
        else:
            logging.warning("Invalid data cannot be loaded to the DataManager.")



__doc__ = """
Bayesian Adaptive Kernel Smoother (BAKS)
BAKS is a method by Ahmandi _[1]_ for estimating progression of firing rate from spiketrain data.
It uses kernel smoothing technique with adaptive bandwidth, based on a Bayesian approach.

References
----------
[1] Ahmadi N, Constandinou TG, Bouganis CS (2018) Estimation of neuronal firing rate using Bayesian Adaptive Kernel Smoother (BAKS). PLOS ONE 13(11): e0206794. https://doi.org/10.1371/journal.pone.0206794
[2] https://github.com/nurahmadi/BAKS
"""
__all__ = ["bayesian_adaptive_kernel_smoother"]

import time

import numpy as np
import scipy.special as sps
from numba import njit

from miv.core.datatype import Spikestamps


def bayesian_adaptive_kernel_smoother(
    spikestamps, probe_time, alpha=4, progress_bar=False
):
    """
    Bayesian Adaptive Kernel Smoother (BAKS)

    Parameters
    ----------
    spiketimes : Spikestamps
        spike event times
    probe_time : array_like
        time at which the firing rate is estimated. Typically, we assume the number of probe_time is much smaller than the number of spikes events.
    alpha : float, optional
        shape parameter, by default 4

    Returns
    -------
    hs : array_like
        adaptive bandwidth (channels, n_time)
    firing_rates : array_like
        estimated firing rate (channels, n_times)
    """
    num_channels = spikestamps.number_of_channels
    firing_rates = np.zeros((num_channels, len(probe_time)))
    hs = np.zeros((num_channels, len(probe_time)))
    for channel in tqdm(
        range(num_channels), desc="Channel: ", disable=not progress_bar
    ):
        spiketimes = np.asarray(spikestamps[channel])
        n_spikes = len(spiketimes)
        if n_spikes == 0:
            continue

        ratio = _numba_ratio_func(probe_time, spiketimes, alpha)
        hs[channel] = (sps.gamma(alpha) / sps.gamma(alpha + 0.5)) * ratio

        firing_rate, firing_rate_for_spike = _numba_firing_rate(
            spiketimes, probe_time, hs[channel]
        )
        firing_rates[channel] = firing_rate
    return hs, firing_rates


@njit(parallel=False)
def _numba_ratio_func(probe_time, spiketimes, alpha):
    # alpha = 1: spike rate contribute up to 1000 sec
    # alpha = 4: spike rate contribute up to 10 sec

    n_spikes = spiketimes.shape[0]
    n_time = probe_time.shape[0]
    sum_numerator = np.zeros(n_time)
    sum_denominator = np.zeros(n_time)

    diff_lim = 10 ** (4.5 / (alpha + 0.5))
    spike_start_indices = np.searchsorted(spiketimes, probe_time - diff_lim)
    spike_end_indices = np.searchsorted(spiketimes, probe_time + diff_lim)

    for j in range(n_time):
        # for i in range(n_spikes):
        _spiketimes = spiketimes[spike_start_indices[j] : spike_end_indices[j]]

        val = (np.square(probe_time[j] - _spiketimes) / 2) ** (-alpha)
        # print(val.shape, val.min(), val.max())
        sum_numerator[j] = val.sum()
        val = (np.square(probe_time[j] - _spiketimes) / 2) ** (-alpha - 0.5)
        sum_denominator[j] = val.sum()

        # for i in range(spike_start_indices[j], spike_end_indices[j]):
        #     val = ((probe_time[j] - spiketimes[i]) ** 2) / 2
        #     sum_numerator[j] += val ** (-alpha)
        #     sum_denominator[j] += val ** (-alpha - 0.5)
    ratio = sum_numerator / (sum_denominator + 1e-14)
    return ratio


@njit(parallel=False)
def _numba_firing_rate(spiketimes, probe_time, h):
    std = 5
    n_spikes = spiketimes.shape[0]
    n_time = probe_time.shape[0]
    firing_rate = np.zeros((n_time, n_spikes))
    # stime = np.searchsorted(spiketimes, probe_time - std * h * 2)
    # etime = np.searchsorted(spiketimes, probe_time + std * h * 2)
    # for j in prange(n_time):

    for j in range(n_time):
        # for i in range(stime[j], etime[j]):
        for i in range(n_spikes):
            firing_rate[j, i] = (1 / (np.sqrt(2 * np.pi) * h[j])) * np.exp(
                -(((probe_time[j] - spiketimes[i]) / (2 * h[j])) ** 2)
            )
    return firing_rate.sum(axis=1), firing_rate


if __name__ == "__main__":
    import time

    from miv.core.datatype import Spikestamps

    from numba import set_num_threads

    set_num_threads(4)

    seed = 0
    np.random.seed(seed)

    def original_imple(a, L=5):
        num = a ** (-L)
        den = a ** (-L - 0.5)
        ratio = num.sum() / den.sum()
        return ratio

    def stable_ratio(a, L=5):
        # Find the maximum element in a to normalize
        a_min = np.min(a)

        # Compute the normalized terms
        normalized_numerator_terms = (a / a_min) ** (-L)
        normalized_denominator_terms = (a / a_min) ** (-L - 0.5)

        # Calculate the sums
        numerator = np.sum(normalized_numerator_terms)
        denominator = np.sum(normalized_denominator_terms)

        # Compute the final stable ratio
        ratio = np.sqrt(a_min) * (numerator / denominator)

        return ratio

    # Test cases
    test_cases = [
        np.array([1.0, 2.0, 3.0]),
        np.array([1e10, 1e5, 1e3]),
        np.array([0.1, 0.5, 0.9]),
        np.array([1.5, 2.5, 3.5]),
        np.array([1e-3, 1e-5, 1e-10]),
        np.geomspace(1e-40, 1e40, 100),
    ]

    # Run the test cases and display results
    for a in test_cases:
        prev_out = original_imple(a)
        stable_out = stable_ratio(a)
        print(prev_out, stable_out)
    sys.exit()

    t = 30
    num_channels = 8
    total_time = 600  # seconds
    N = 80 * total_time  # Number of spikes
    n_evaluation_points = int(total_time)
    evaluation_points = np.linspace(0, total_time, n_evaluation_points)

    spikestamps = [
        np.sort(np.random.random(N)) * total_time for _ in range(num_channels)
    ]
    spikestamps = Spikestamps(spikestamps)

    stime = time.time()
    alpha = 4.0
    hs, firing_rates = bayesian_adaptive_kernel_smoother(
        spikestamps, evaluation_points, alpha=alpha
    )
    etime = time.time()
    # print(alpha, beta)

    std = hs[0][0] * 3

    # print(hs)
    # print(firing_rates)

    # fig, (ax1, ax2, ax3) = plt.subplots(3, 1, sharex=True)
    # # plt.plot(evaluation_points, firing_rates[0])
    # ax1.eventplot(spikestamps[0], color="k")
    # ax1.axvline(t, color="r", linestyle="--")
    # ax1.axvspan(t - std, t + std, color="r", alpha=0.5)

    # ax2.plot(spikestamps[0], firing_rate_for_spike[0][0])
    # ax2.axvline(t, color="r", linestyle="--")
    # ax2.axvspan(t - std, t + std, color="r", alpha=0.5)

    # ax3.plot(spikestamps[0], np.cumsum(firing_rate_for_spike[0][0]))
    # ax3.axvline(t, color="r", linestyle="--")
    # ax3.axvspan(t - std, t + std, color="r", alpha=0.5)

    # plt.figure()
    # plt.plot(evaluation_points, firing_rates[0])

    # plt.show()

    print(f"Elapsed time: {etime - stime:.4f} seconds")



__doc__ = """
Signal
======

.. autoclass:: Signal
   :members:

"""

__all__ = ["Signal"]

from typing import cast

import pickle
from dataclasses import dataclass

import numpy as np

from miv.core.datatype.mixin_colapsable import ConcatenateMixin
from miv.core.operator.operator import DataNodeMixin
from miv.core.operator.policy import SupportMultiprocessing
from miv.typing import SignalType, SpikestampsType


@dataclass
class Signal(SupportMultiprocessing, DataNodeMixin, ConcatenateMixin):
    """
    Contiguous array of raw signal type.

    [signal length, number of channels]
    """

    # This choice of axis is mainly due to the memory storage structure.
    _CHANNELAXIS = 1
    _SIGNALAXIS = 0

    data: SignalType
    timestamps: SpikestampsType
    rate: float = 30_000

    def __post_init__(self) -> None:
        super().__init__()
        self.data = np.asarray(self.data)
        assert len(self.data.shape) == 2, "Signal must be 2D array"

    @property
    def number_of_channels(self) -> int:
        """Number of channels in the signal."""
        return int(self.data.shape[self._CHANNELAXIS])

    def __getitem__(self, i: int) -> SignalType:
        return self.data[:, i]

    def select(self, indices: tuple[int, ...]) -> "Signal":
        """Select channels by indices."""
        return Signal(self.data[:, indices], self.timestamps, self.rate)

    def get_start_time(self) -> float:
        """Get the start time of the signal."""
        return cast(float, self.timestamps.min())

    def get_end_time(self) -> float:
        """Get the end time of the signal."""
        return cast(float, self.timestamps.max())

    @property
    def shape(self) -> tuple[int, int]:
        """Shape of the signal."""
        return cast(tuple[int, int], self.data.shape)

    def append(self, value: np.ndarray) -> None:
        """Append a channels to the end of the existing signal."""
        assert value.shape[self._SIGNALAXIS] == self.data.shape[self._SIGNALAXIS]
        self.data = np.append(self.data, value, axis=self._CHANNELAXIS)

    def extend_signal(self, data: np.ndarray, time: SpikestampsType) -> None:
        """Append a signal to the end of the existing signal."""
        assert data.shape[self._SIGNALAXIS] == time.shape[0]
        assert data.shape[self._CHANNELAXIS] == self.data.shape[self._CHANNELAXIS], (
            "Signal must have same number of channels"
        )
        self.data = np.append(self.data, data, axis=self._SIGNALAXIS)
        self.timestamps = np.append(self.timestamps, time)

    def extend(self, value: "Signal") -> None:
        """Append a signal to the end of the existing signal."""
        self.extend_signal(value.data, value.timestamps)

    def prepend_signal(self, data: np.ndarray, time: SpikestampsType) -> None:
        """Prepend a signal to the end of the existing signal."""
        assert data.shape[self._SIGNALAXIS] == time.shape[0], (
            "Time and signal must have same length"
        )
        assert data.shape[self._CHANNELAXIS] == self.data.shape[self._CHANNELAXIS], (
            "Signal must have same number of channels"
        )
        self.data = np.append(data, self.data, axis=self._SIGNALAXIS)
        self.timestamps = np.append(self.timestamps, time)

    def save(self, path: str) -> None:
        """Save signal to file."""
        with open(path, "wb") as f:
            pickle.dump(self, f)

    @classmethod
    def load(cls, path: str) -> "Signal":
        """Load signal from file."""
        with open(path, "rb") as f:
            return cast(Signal, pickle.load(f))



import pathlib
import shutil

import click


@click.command()
@click.option(
    "--path",
    "-p",
    default=".",
    type=click.Path(exists=True),
    help="Path to clean cache files in.",
)
@click.option(
    "--dry",
    "-d",
    is_flag=True,
    show_default=True,
    default=False,
    help="Dry run. No files will be deleted.",
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    show_default=True,
    default=False,
    help="Verbose output.",
)
def clean_cache(path, dry, verbose):
    """
    Clean all cache files, recursively, in the given path.
    """
    if dry:
        print("Dry run. No folders will be actually deleted.")
    pattern = pathlib.Path(path).glob("**/.cache")
    for path in pattern:
        if verbose:
            print(f"Removing: {path}", flush=True)
        if not dry:
            shutil.rmtree(path)



__all__ = ["MEAGeometryProtocol"]

from typing import Protocol

import numpy as np

from miv.core.protocol import _Jsonable
from miv.core.operator.protocol import _Chainable


class MEAGeometryProtocol(_Jsonable, _Chainable, Protocol):
    @property
    def coordinates(self) -> np.ndarray:
        """Return coordinates of MEA electrodes location"""
        ...

    def get_xy(self, idx: int) -> tuple[float, float]:
        """Given node index, return xy coordinate"""
        ...

    def get_ixiy(self, idx: int) -> tuple[int, int]:
        """Given node index, return coordinate index"""
        ...

    def save(self, path: str) -> None: ...

    def load(self, path: str) -> None: ...

    def view(self):
        """Simplified view of MEA orientation"""
        ...

    def map_data(self, vector: np.ndarray, missing_value: float) -> np.ndarray:
        """Map data (1-D array) to MEA (2-D array or N-D array)"""
        ...



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    import mpi4py


def task_index_split(comm: mpi4py.MPI.Comm, num_tasks: int) -> list[int]:
    # TODO documentation
    # ex) split [1,2,3,4] --> [1,2], [3,4]

    rank = comm.Get_rank()
    size = comm.Get_size()

    tasks = np.array_split(np.arange(num_tasks), size)[rank].tolist()
    return tasks



def pbar(iterable, logger, step=25):
    """
    Every step%, log the progress of the iterable

    How to use:
    >>> from miv.utils.progress_logging import pbar
    >>> import logging
    >>> logger = logging.getLogger(__name__)
    >>> for i in pbar(range(100), logger, step=25):
    ...     pass
    complete: 25%
    complete: 50%
    complete: 75%
    complete: 100%
    """
    assert 0 < step <= 100, "step must be in (0, 100]"
    total = len(iterable)
    for i, item in enumerate(iterable):
        yield item
        if (i + 1) % (total // step) == 0:
            logger.info(f"complete: {((i + 1) / total) * 100:.0f}%")
    logger.info("complete: 100%")



__doc__ = """Protocol for Data objects."""
__all__ = ["DataProtocol"]

from typing import (
    Any,
    Protocol,
)
from collections.abc import Generator


from miv.core.datatype.signal import Signal
from miv.core.datatype.spikestamps import Spikestamps


class DataProtocol(Protocol):
    """Behavior definition for a single experimental data handler."""

    def __init__(self, data_path: str, tag: str = "data"): ...

    @property
    def data_path(self) -> str: ...

    @property
    def tag(self) -> str: ...

    def load(self, *args: Any) -> Generator[Signal] | Spikestamps | Signal:
        """Iterator to load data fragmentally. Use to load large file size data."""
        ...

    def check_path_validity(self) -> bool:
        """Check if necessary files exist in the directory."""
        ...



__doc__ = """
Module for filtering spiketrain based on waveform statistical properties.

Motivated by _[1].

[1]. Toosi R., et al. An automatic spike sorting algorithm based on adaptive spike detection and a mixture of skew-t distributions (2021).
"""
__all__ = ["WaveformStatisticalFilter"]


import csv
import os
import pathlib
from dataclasses import dataclass

import matplotlib.pyplot as plt
import numpy as np

from miv.core.datatype import Signal, Spikestamps
from miv.core.operator.operator import OperatorMixin
from miv.visualization.event import plot_spiketrain_raster


@dataclass
class WaveformStatisticalFilter(OperatorMixin):
    """
    Filter spike train based on waveform statistical properties.

    Mainly, the waveform that has a mean larger than 1.0 and standard deviation larger than 3.0.

    .. code::

        >>> filtered_signal = ButterBandpass(300, 3000, order=4)
        >>> spiketrains = ThresholdCutoff(cutoff=4.0)
        >>> waveform = ExtractWaveforms()

        >>> filtered_signal >> spiketrains
        >>> filtered_signal >> waveform; spiketrains >> waveform
        >>> waveform >> statistical_filter; spiketrains >> statistical_filter

        >>> Pipeline(statistical_filter).run()

    Parameters
    ----------
    max_mean : float
        The maximum mean of the waveform. (default: 1.0)
    max_std : float
        The maximum standard deviation of the waveform. (default: 3.0)
    """

    max_mean: float = 1.0
    max_std: float = 10.0

    tag: str = "waveform statistical filter"

    def __post_init__(self):
        super().__init__()

    def __call__(
        self, waveforms: dict[int, Signal], spiketrains: Spikestamps
    ) -> Spikestamps:
        """
        Parameters
        ----------
        waveforms : Dict[int, Signal]
            The waveforms of each channel.
        spiketrains : Spikestamps
            The spiketrains of each channel.

        Returns
        -------
        Spikestamps
            The filtered spiketrains of each channel.
        """

        filtered_spiketrains = Spikestamps()
        total_spikecounts = spiketrains.get_count()
        log = []
        for channel in range(spiketrains.number_of_channels):
            if channel not in waveforms:
                filtered_spiketrains.append(np.array([]))
                continue
            waveform = waveforms[channel]
            spiketrain = spiketrains[channel]
            total_spikecount = total_spikecounts[channel]

            spike_mean = np.mean(waveform.data, axis=waveform._SIGNALAXIS)
            spike_std = np.std(waveform.data, axis=waveform._SIGNALAXIS)

            indices = (
                (spike_std > self.max_std)
                | (spike_mean > self.max_mean)
                | (spike_mean < -self.max_mean)
            )
            number_of_outliers = indices.sum()

            filtered_spiketrains.append(np.asarray(spiketrain)[~indices])
            log.append(
                (
                    channel,
                    total_spikecount,
                    number_of_outliers,
                    number_of_outliers / total_spikecount,
                )
            )

        self._log = log  # FIXME: Maybe better way to pass down the result

        return filtered_spiketrains

    def after_run_save_log(self, output, inputs, *args, **kwargs):
        log = self._log
        # Save log in csv
        savepath = os.path.join(self.analysis_path, "filtered_statistics.csv")
        with open(savepath, "w", newline="") as csvfile:
            writer = csv.writer(csvfile)
            writer.writerow(
                ["channel", "total spikecount", "number of filtered outliers", "ratio"]
            )
            for row in log:
                writer.writerow(row)
        return output

    def plot_spiketrain(
        self,
        spikestamps,
        inputs,
        show: bool = False,
        save_path: pathlib.Path | None = None,
    ) -> plt.Axes:
        """
        Plot spike train in raster
        """
        t0 = spikestamps.get_first_spikestamp()
        tf = spikestamps.get_last_spikestamp()

        # TODO: REFACTOR. Make single plot, and change xlim
        term = 60
        n_terms = int(np.ceil((tf - t0) / term))
        if n_terms == 0:
            # TODO: Warning message
            return None
        for idx in range(n_terms):
            fig, ax = plot_spiketrain_raster(
                spikestamps, idx * term + t0, min((idx + 1) * term + t0, tf)
            )
            if save_path is not None:
                plt.savefig(os.path.join(save_path, f"spiketrain_raster_{idx:03d}.png"))
            if not show:
                plt.close("all")
        if show:
            plt.show()
            plt.close("all")
        return ax



__doc__ = """"""

from typing import TYPE_CHECKING, Any, Protocol

import pathlib

import matplotlib.pyplot as plt

from miv.core.operator.callback import (
    get_methods_from_feature_classes_by_startswith_str,
)

if TYPE_CHECKING:
    from miv.core.datatype import DataTypes


class _GeneratorCallback(Protocol):
    _done_flag_generator_plot: bool
    _done_flag_firstiter_plot: bool

    def _callback_generator_plot(
        self,
        iter_index: int,
        output: "DataTypes",
        inputs: tuple["DataTypes", ...] | None = None,
        show: bool = False,
        save_path: str | pathlib.Path | None = None,
    ) -> None: ...

    def _callback_firstiter_plot(
        self,
        output: "DataTypes",
        inputs: tuple["DataTypes", ...] | None = None,
        show: bool = False,
        save_path: str | pathlib.Path | None = None,
    ) -> None: ...


class GeneratorCallbackMixin:
    """
    Additional methods for generator-to-generator operator.

    `generator_plot` method plots during each iteration of generator.
    The function take `show` and `save_path` arguments similar to `plot` method.
    """

    def __init__(self) -> None:
        super().__init__()

        self._done_flag_generator_plot = False
        self._done_flag_firstiter_plot = False

    def reset_callbacks(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._done_flag_generator_plot = getattr(kwargs, "plot", False)
        self._done_flag_firstiter_plot = getattr(kwargs, "plot", False)

    def _callback_generator_plot(
        self,
        iter_index: int,
        output: "DataTypes",
        inputs: tuple["DataTypes", ...] | None = None,
        show: bool = False,
        save_path: bool | str | pathlib.Path | None = None,
    ) -> None:
        if self._done_flag_generator_plot:
            return

        plotters_for_generator_out = get_methods_from_feature_classes_by_startswith_str(
            self, "generator_plot_"
        )
        for plotter in plotters_for_generator_out:
            plotter(
                output,
                inputs,
                show=show,
                save_path=save_path,
                index=iter_index,
            )
        plt.close("all")

    def _callback_firstiter_plot(
        self,
        output: "DataTypes",
        inputs: tuple["DataTypes", ...] | None = None,
        show: bool = False,
        save_path: str | pathlib.Path | None = None,
    ) -> None:
        if self._done_flag_firstiter_plot:
            return

        plotters_for_generator_out = get_methods_from_feature_classes_by_startswith_str(
            self, "firstiter_plot_"
        )

        for plotter in plotters_for_generator_out:
            plotter(
                output,
                inputs,
                show=show,
                save_path=save_path,
            )
        plt.close("all")



__doc__ = """
Events
======

.. autoclass:: Events
   :members:

"""

__all__ = ["Events"]

from typing import cast


import numpy as np
import quantities as pq

from miv.core.datatype.mixin_colapsable import ConcatenateMixin
from miv.core.datatype.signal import Signal
from miv.core.operator.operator import DataNodeMixin


class Events(ConcatenateMixin, DataNodeMixin):
    """
    A list of events.

    Comply with `Extendable` protocols.
    """

    def __init__(self, data: list[float] | None = None) -> None:
        super().__init__()
        self.data = np.asarray(data) if data is not None else np.array([])

    def append(self, item: float) -> None:
        self.data = np.append(self.data, item)

    def extend(self, other: "Events") -> None:
        self.data = np.append(self.data, other.data)

    def __len__(self) -> int:
        return len(self.data)

    def get_last_event(self) -> float:
        """Return timestamps of the last event"""
        return cast(float, max(self.data))

    def get_first_event(self) -> float:
        """Return timestamps of the first event"""
        return cast(float, min(self.data))

    def get_view(self, t_start: float, t_end: float) -> "Events":
        """Truncate array and only includes spikestamps between t_start and t_end."""
        return Events(sorted(filter(lambda x: t_start <= x <= t_end, self.data)))

    def binning(
        self,
        bin_size: float | pq.Quantity = 0.001,
        t_start: float | None = None,
        t_end: float | None = None,
        return_count: bool = False,
    ) -> Signal:
        """
        Forms a binned events

        Parameters
        ----------
        bin_size : float | pq.Quantity
            bin size in the unit of time.
        return_count : bool
            If set to true, return the bin count. (default=False)
        """

        if isinstance(bin_size, pq.Quantity):
            bin_size = bin_size.rescale(pq.s).magnitude
        assert bin_size > 0, "bin size should be greater than 0"

        t_start = self.get_first_event() if t_start is None else t_start
        t_end = self.get_last_event() if t_end is None else t_end
        n_bins = int(np.ceil((t_end - t_start) / bin_size))
        time = t_start + (np.arange(n_bins + 1) * bin_size)

        signal = Signal(
            data=np.zeros(
                [time.shape[0] - 1, 1],
                dtype=np.int_ if return_count else np.bool_,
            ),
            timestamps=time[:-1],
            rate=1.0 / bin_size,
        )

        # TODO: Make separate free function for this binning process
        bins = np.digitize(self.data, time)
        bincount = np.bincount(bins, minlength=n_bins + 2)[1:-1]
        if return_count:
            bin_events = bincount
        else:
            bin_events = (bincount != 0).astype(np.bool_)
        signal.data[:, 0] = bin_events

        return signal



__all__ = [
    "SpikeDetectionProtocol",
    "SpikeFeatureExtractionProtocol",
    "UnsupervisedFeatureClusteringProtocol",
]

from typing import Any, Protocol

import numpy as np

from miv.typing import SpikestampsType


class SpikeDetectionProtocol(Protocol):
    def __call__(self, *args: Any, **kwargs: Any) -> SpikestampsType: ...


# TODO: Behavior is clear, but not sure what the name should be
# class SpikeSortingProtocol(Protocol):
#    def __call__(
#        self, signals: Iterable[SignalType]
#    ) -> np.ndarray:
#        ...
#
#    def __repr__(self) -> str:
#        ...


class SpikeFeatureExtractionProtocol(Protocol):
    """ex) wavelet transform, PCA, etc."""

    def __repr__(self) -> str: ...


class UnsupervisedFeatureClusteringProtocol(Protocol):
    def __repr__(self) -> str: ...

    def fit(self, X: np.ndarray): ...

    def predict(self, X: np.ndarray) -> np.ndarray: ...



__doc__ = """
Module for extracting each spike waveform and visualize.
"""
__all__ = ["ExtractWaveforms", "WaveformAverage"]

import inspect
import os
import pathlib
from collections.abc import Generator
from dataclasses import dataclass
from typing import Any

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import quantities as pq
from matplotlib import cm
from tqdm import tqdm

from miv.core.datatype import Signal, Spikestamps
from miv.core.operator.operator import OperatorMixin
from miv.core.operator.wrapper import cache_call
from miv.mea import MEAGeometryProtocol


@dataclass
class ExtractWaveforms(OperatorMixin):
    """
    Extract spike waveforms as signal cutouts around each spike index as a spikes x samples numpy array

    Parameters
    ----------
    channels : Optional[List[int]]
        Interested channels. If None, extract from all channels. (default: None)
    pre : pq.Quantity | float
        The duration of the cutout before the spike in seconds. (default: 0.001 s)
    post : pq.Quantity | float
        The duration of the cutout after the spike in seconds. (default: 0.002 s)
    plot_n_spikes : Optional[int]
        The number of cutouts to plot. None to plot all. (default: 100)
    """

    channels: list[int] | None = None
    pre: float = 0.001
    post: float = 0.002
    plot_n_spikes: int = 100
    tag: str = "extract waveform"

    progress_bar: bool = False

    def __post_init__(self):
        super().__init__()
        self.cacher.policy = "ON"

    @cache_call
    def __call__(
        self, signal: Generator[Signal, None, None], spikestamps: Spikestamps
    ) -> dict[int, Signal]:
        """__call__

        Parameters
        ----------
        signal : Generator[Signal, None, None]
            The signal to extract the event feature
        spikestamps : Spikestamps
            The sample index of all spikes as a 1-dim numpy array

        Returns
        -------
        Stack of spike cutout: List[np.ndarray]
            Return stacks of spike cutout; shape(n_spikes, width).

        """
        if isinstance(self.pre, pq.Quantity):
            pre = self.pre.rescale(pq.s).magnitude
        else:
            pre = self.pre
        if isinstance(self.post, pq.Quantity):
            post = self.post.rescale(pq.s).magnitude
        else:
            post = self.post
        if not inspect.isgenerator(signal):
            signal = iter([signal])

        def signal_pair_generator():
            ps = next(signal)
            for s in signal:
                yield ps, s
                ps = s
            yield ps, None

        waveforms = {}
        previous_sig = None
        for sig, next_sig in tqdm(
            signal_pair_generator(), desc="Iterate signal stream"
        ):
            sampling_rate = sig.rate
            num_channels = sig.number_of_channels
            channels = range(num_channels) if self.channels is None else self.channels
            pre_idx = int(pre * sampling_rate)
            post_idx = int(post * sampling_rate)
            assert pre_idx + post_idx > 0, (
                "Set larger pre/post duration. pre+post duration must be more than 1/sampling_rate."
            )
            spikestamps_view = spikestamps.get_view(
                sig.get_start_time(), sig.get_end_time()
            )
            for ch in channels:
                # Padding signal
                spikestamp = spikestamps_view[ch]
                if len(spikestamp) == 0:
                    continue
                padded_signal = np.pad(
                    sig.data[:, ch], ((pre_idx, post_idx),), constant_values=0
                )
                if (
                    previous_sig is not None
                ):  # TODO: channel axis might change in the future
                    padded_signal[:pre_idx] = previous_sig.data[-pre_idx:, ch]
                if next_sig is not None:
                    padded_signal[-post_idx:] = next_sig.data[:post_idx, ch]
                cutout = []
                for idx, time in enumerate(spikestamp):
                    index = int(round((time - sig.get_start_time()) * sampling_rate))
                    cutout.append(padded_signal[index : (index + post_idx + pre_idx)])
                cutout = np.stack(cutout, axis=1)
                if ch not in waveforms:
                    waveforms[ch] = Signal(
                        data=cutout,
                        timestamps=(
                            np.arange(pre_idx + post_idx).astype(np.float64) - pre
                        )
                        / sampling_rate,
                        rate=sampling_rate,
                    )
                else:
                    waveforms[ch].append(cutout)
            previous_sig = sig

        return waveforms

    def plot_waveforms(
        self,
        waveforms: dict[int, Signal],
        inputs,
        show: bool = False,
        save_path: pathlib.Path | None = None,
        plot_kwargs: dict[Any, Any] = None,
    ):
        """
        Plot an overlay of spike waveforms

        Parameters
        ----------
        waveforms : np.ndarray
            A spikes x samples array of cutouts waveforms.
        plot_kwargs : Dict[Any, Any]
            Addtional keyword-arguments for matplotlib.pyplot.plot.
        """
        for ch, signal in waveforms.items():
            cutout = signal.data
            num_cutout = signal.number_of_channels
            if self.plot_n_spikes is None:
                self.plot_n_spikes = num_cutout
            plot_n_spikes = min(self.plot_n_spikes, num_cutout)
            if num_cutout == 0:
                continue

            if not plot_kwargs:
                plot_kwargs = {"alpha": 0.3, "linewidth": 1, "color": "b"}

            time = signal.timestamps * pq.s
            time = time.rescale(pq.ms).magnitude

            fig, ax = plt.subplots(figsize=(8, 5))
            for i in range(plot_n_spikes):
                arr = cutout[:, i]
                arr[np.isnan(arr)] = 0
                ax.plot(
                    time,
                    arr,
                    **plot_kwargs,
                )
            ax.set_xlabel("Time (ms)")
            ax.set_ylabel("Voltage (microV)")
            ax.set_title(f"Spike Cutouts (channel {ch})")
            if save_path:
                plt.savefig(os.path.join(save_path, f"spike_cutouts_ch{ch:03}.png"))
            plt.close(fig)
        plt.close("all")


@dataclass
class WaveformAverage(OperatorMixin):
    """
    Plot the average waveform of each channel
    """

    tag: str = "waveform average"

    def __post_init__(self):
        super().__init__()

    @cache_call
    def __call__(self, waveforms: dict[int, Signal], mea: MEAGeometryProtocol):
        mean_waveform = {}
        std_waveform = {}
        for channel, cutout in waveforms.items():
            mean_waveform[channel] = cutout.data.mean(axis=cutout._CHANNELAXIS)
            std_waveform[channel] = cutout.data.std(axis=cutout._CHANNELAXIS) / np.sqrt(
                cutout.number_of_channels
            )
        return mean_waveform, std_waveform

    def plot_averaged_waveform(self, output, inputs, show=False, save_path=None):
        waveforms, mea = self.receive()
        mean_waveform, std_waveform = output
        nrow, ncol = mea.nrow, mea.ncol

        fig, axes = plt.subplots(nrow, ncol, figsize=(ncol * 6, nrow * 4))
        for channel, cutout in waveforms.items():
            idx = mea.get_ixiy(channel)
            if idx is None:
                continue
            ax = axes[idx[0], idx[1]]
            cutout_mean = mean_waveform[channel]
            cutout_stdmn = std_waveform[channel]
            ax.fill_between(
                cutout.timestamps,
                cutout_mean - 2 * cutout_stdmn,
                cutout_mean + 2 * cutout_stdmn,
                alpha=0.5,
            )
            ax.plot(cutout.timestamps, cutout_mean, "k-")
            ax.set_title(f"channel {channel}")
        # Bottom row
        for i in range(ncol):
            axes[-1, i].set_xlabel("time (s)")

        # Left row
        for i in range(nrow):
            axes[i, 0].set_ylabel("Voltage (microV)")

        plt.suptitle("Waveform Average Plot")

        os.makedirs(self.analysis_path, exist_ok=True)
        plt.savefig(os.path.join(self.analysis_path, "waveform_average.png"))
        plt.close(fig)

    def plot_waveform_pregression_3d(self, output, inputs, show=False, save_path=None):
        waveforms, mea = self.receive()
        mean_waveform, std_waveform = output
        nrow, ncol = mea.nrow, mea.ncol

        fig, axes = plt.subplots(
            nrow, ncol, figsize=(ncol * 6, nrow * 4), subplot_kw=dict(projection="3d")
        )
        for channel, cutout in waveforms.items():
            idx = mea.get_ixiy(channel)
            if idx is None:
                continue
            ax = axes[idx[0], idx[1]]
            time = cutout.timestamps  # X
            iter_number = np.arange(cutout.number_of_channels)  # Y
            X, Y = np.meshgrid(iter_number, time)
            zmax = (mean_waveform[channel] + std_waveform[channel] * 2.5).max()
            zmin = (mean_waveform[channel] - std_waveform[channel] * 2.5).min()
            divnorm = matplotlib.colors.TwoSlopeNorm(vmin=zmin, vcenter=0, vmax=zmax)
            surf = axes[idx[0], idx[1]].plot_surface(
                X, Y, cutout.data, cmap=cm.coolwarm, norm=divnorm
            )
            ax.set_zlim(zmin, zmax)
            fig.colorbar(surf, shrink=0.5, ax=ax)

            ax.set_title(f"channel {channel}")
            ax.set_xlabel("iteration")
            ax.set_ylabel("time (s)")
            ax.set_zlabel("voltage (microV)")
        plt.suptitle("Waveform Transient Plot")

        os.makedirs(self.analysis_path, exist_ok=True)
        plt.savefig(os.path.join(self.analysis_path, "waveform_transient.png"))
        plt.close(fig)



__doc__ = """Multi-channel signal and spike rendering for each MEA channels"""
__all__ = ["ReplayRecording"]


import inspect
import logging
import os
import shutil
from dataclasses import dataclass

from matplotlib import gridspec
import numpy as np
from matplotlib import pyplot as plt
from tqdm import tqdm

from miv.core.datatype import Signal, Spikestamps
from miv.core.operator.operator import OperatorMixin
from miv.core.operator.policy import StrictMPIRunner
from miv.mea.protocol import MEAGeometryProtocol
from miv.utils.cli import command_run


@dataclass
class ReplayRecording(OperatorMixin):
    """
    Collective rendering of electrophyiological signals from MEA channels.
    """

    window_size: float = 3.0  # sec. TODO: refactor
    play_speed: float = 1.0

    tag: str = "experiment replay"

    progress_bar: bool = False

    fps: int = 30
    dpi: int = 200

    def __post_init__(self):
        super().__init__()
        self.runner = StrictMPIRunner()

    def __call__(
        self,
        signals: Signal,
        signals2: Signal,
        spikestamps: Spikestamps,
        cutouts: dict[int, Signal],
        mea: MEAGeometryProtocol,
    ):  # pragma: no cover
        comm = self.runner.comm
        rank = self.runner.get_rank()
        size = self.runner.get_size()

        assert inspect.isgenerator(signals)
        assert inspect.isgenerator(signals2)

        for vidx, (signal, signal2) in enumerate(zip(signals, signals2, strict=False)):
            spikestamps_view = spikestamps.get_view(
                signal.get_start_time(), signal.get_end_time()
            )
            rate = signal.rate

            n_steps_in_window = int(self.window_size * rate)
            interval = int(rate // self.fps * self.play_speed)
            n_steps = int((signal.timestamps.size - n_steps_in_window) // interval)

            # Distribute tasks
            if size > n_steps:
                if rank >= n_steps:
                    logging.warning(f"rank {rank} is idle")
                    continue
                size = n_steps
            tasks = np.arange(n_steps)
            mytask = np.array_split(tasks, size)[rank]

            # Output Images
            plt.rcParams.update({"font.size": 10})

            images_path = os.path.join(self.analysis_path, f"render_{vidx=}")
            if self.runner.is_root():
                os.makedirs(self.analysis_path, exist_ok=True)
                if os.path.exists(images_path):
                    shutil.rmtree(images_path)
                os.makedirs(images_path, exist_ok=True)
            comm.barrier()

            fig = plt.figure(figsize=(16, 20))
            if self.runner.is_root():
                pbar = tqdm(
                    total=len(mytask),
                    desc="Rendering (rank 0)",
                    disable=not self.progress_bar,
                )
            for step in mytask:
                fig.clf()
                outer = gridspec.GridSpec(mea.nrow, mea.ncol, wspace=0.2, hspace=0.2)

                sindex = int(step * interval)
                time = signal.timestamps[sindex : sindex + n_steps_in_window]

                for channel in range(signal.number_of_channels):
                    if channel not in cutouts:
                        continue
                    if channel not in mea.grid:
                        continue

                    loc = mea.get_ixiy(channel)
                    mea_index = loc[0] * mea.ncol + loc[1]
                    inner = gridspec.GridSpecFromSubplotSpec(
                        3,
                        1,
                        subplot_spec=outer[mea_index],
                        wspace=0.1,
                        hspace=0.1,
                        height_ratios=[0.4, 0.4, 0.2],
                    )

                    # Plot signal on the top
                    ax1 = fig.add_subplot(inner[0])
                    ax1.plot(
                        time,
                        signal.data[sindex : sindex + n_steps_in_window, channel],
                    )
                    ax1.set_title(f"Channel {channel}")
                    med = np.median(np.abs(signal.data[:, channel])) * 15
                    y_min = -med
                    y_max = med
                    ax1.set_ylim(y_min, y_max)

                    # Plot signal on the middle
                    ax2 = fig.add_subplot(inner[1], sharex=ax1)
                    ax2.plot(
                        time,
                        signal2.data[sindex : sindex + n_steps_in_window, channel],
                    )
                    med = np.median(np.abs(signal2.data[:, channel])) * 15
                    y_min = -med
                    y_max = med
                    ax2.set_ylim(y_min, y_max)

                    # Plot spikestamps on the bottom
                    ax3 = fig.add_subplot(inner[2], sharex=ax1)
                    ax3.eventplot(
                        spikestamps_view.get_view(time.min(), time.max())[channel]
                    )
                    ax3.set_yticks([])

                    if loc[0] == mea.nrow - 1:
                        ax3.set_xlabel("time (sec)")
                    else:
                        ax3.set_xticks([])
                        ax3.set_xticks([], minor=True)

                fig.tight_layout()
                plt.savefig(
                    os.path.join(images_path, f"{step:05d}.png"),
                    dpi=self.dpi,
                )
                if self.runner.is_root():
                    pbar.update(1)
            if self.runner.is_root():
                pbar.close()
            plt.close(plt.gcf())
            comm.barrier()

            if self.runner.is_root():
                # Concatenate images using ffmpeg
                video_name = os.path.join(self.analysis_path, f"render_{vidx=}.mp4")
                ffmpeg_path = shutil.which("ffmpeg")
                if ffmpeg_path is None:
                    logging.warning("ffmpeg not found")
                else:
                    # "-threads",
                    # f"{mp.cpu_count()}",
                    cmd = [
                        f"{ffmpeg_path}",
                        "-r",
                        f"{self.fps}",
                        "-i",
                        f"{images_path}/%05d.png",
                        "-b:v",
                        "90M",
                        "-vcodec",
                        "mpeg4",
                        f"{video_name}",
                    ]
                    command_run(cmd, logger=self.logger)

                # if os.path.exists(images_path):
                #    shutil.rmtree(images_path)
            comm.barrier()



__all__ = ["command_run"]

import subprocess


def command_run(cmd, logger=None, exception="skip"):  # pragma: no cover
    args = cmd
    output = subprocess.run(args, capture_output=True, check=False)
    if logger is not None:
        logger.info(f"Running: {cmd=}")
        logger.info(f"{output.returncode=}")
        if output.returncode == 0:
            logger.info(f"{output.stdout.decode('utf-8')=}")
        else:
            logger.info(f"{output.stderr.decode('utf-8')=}")
        logger.info("Done")
    if output.returncode != 0:
        if exception == "raise":
            raise RuntimeError(f"Error running {cmd=}\n{output.stderr.decode('utf-8')}")
        elif exception == "skip":
            if logger is not None:
                logger.warning(
                    f"CLI error running {cmd=}\n{output.stderr.decode('utf-8')}"
                )
    return output.returncode



__all__ = ["zip_results"]


import pathlib
import zipfile

import click


def is_path_valid(
    path: pathlib.Path, ignore_directory: list[str], ignore_extension: list[str]
):
    if path.is_file():
        if ignore_extension and path.suffix in ignore_extension:
            return False
    elif not ignore_directory:
        return True

    if ignore_directory:
        for directory in path.as_posix().split("/"):
            if directory in ignore_directory:
                return False

    return True


def zip_directory_recursively(
    path: pathlib.Path,
    root_dir: pathlib.Path,
    zipfile_handle=None,
    ignore_directory=None,
    ignore_extension=None,
    verbose: bool = False,
):
    ignore_directory = ignore_directory or []
    ignore_extension = ignore_extension or []

    path = pathlib.Path(path)
    root_dir = pathlib.Path(root_dir)

    if path.is_file():
        if is_path_valid(path, ignore_directory, ignore_extension):
            if zipfile_handle is not None:  # dry run
                zipfile_handle.write(path, path.relative_to(root_dir))
        return

    if verbose:
        print(f"    {path.as_posix()}")
    for sub_dir in path.iterdir():
        if not is_path_valid(sub_dir, ignore_directory, ignore_extension):
            if verbose:
                print(f"    ignored: {sub_dir.as_posix()}")
            continue

        zip_directory_recursively(
            sub_dir,
            root_dir,
            zipfile_handle,
            ignore_directory,
            ignore_extension,
            verbose,
        )


@click.command()
@click.option(
    "--path",
    "-p",
    default=".",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    help="Result path to compress",
)
@click.option(
    "--output-file",
    "-o",
    default="results.zip",
    type=click.Path(),
    help="Output file name",
)
@click.option(
    "--include-cache",
    is_flag=True,
    show_default=True,
    default=False,
    help="Include .cache folders.",
)
@click.option(
    "--dry",
    "-d",
    is_flag=True,
    show_default=True,
    default=False,
    help="Dry run. No files will be created.",
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    show_default=True,
    default=False,
    help="Verbose output.",
)
def zip_results(path, output_file, include_cache, dry, verbose):
    """
    Zip the result path.
    """
    if dry:
        print("Dry run. No files will be created.")
    if verbose:
        print(f"Compressing: {path}")
        print(f"Output file: {output_file}")
        if include_cache:
            print("Including .cache folders.")

    ignore_dir = [".cache"]
    ignore_ext = []
    if dry:
        zip_handle = None
    else:
        zip_handle = zipfile.ZipFile(output_file, "w")

    try:
        zip_directory_recursively(
            path, path, zip_handle, ignore_dir, ignore_ext, verbose
        )
    finally:
        if not dry:
            zip_handle.close()

    if verbose:
        print("Compression done.")



__doc__ = """
Specification of the behaviors for Operator modules.
"""

from typing import Protocol, Any, TypeVar
from collections.abc import Iterator

import pathlib
from abc import abstractmethod

from .policy import _RunnerProtocol
from .cachable import _CacherProtocol, CACHE_POLICY
from ..protocol import _Loggable
from miv.core.datatype import DataTypes

C = TypeVar("C", bound="_Chainable")


class _Runnable(Protocol):
    """
    A protocol for a runner policy.
    """

    @property
    def runner(
        self,
    ) -> _RunnerProtocol: ...

    @abstractmethod
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...


class _Cachable(Protocol):
    """
    A protocol for cachable behavior.
    """

    cacher: _CacherProtocol

    def set_caching_policy(self, policy: CACHE_POLICY) -> None: ...


class _Chainable(Protocol[C]):
    """
    Defines the behavior for chaining operator modules:
    - Forward direction defines execution order
    - Backward direction defines dependency order
    """

    def append_upstream(self, node: C) -> None: ...

    def append_downstream(self, node: C) -> None: ...

    def disconnect_upstream(self, node: C) -> None: ...

    def disconnect_downstream(self, node: C) -> None: ...

    def iterate_upstream(self) -> Iterator[C]: ...

    def iterate_downstream(self) -> Iterator[C]: ...


class _Node(
    _Loggable,
    _Runnable,
    _Chainable,
    _Cachable,
    Protocol,
):
    """
    Protocol defining the complete behavior of an Operator node.
    Each protocol aspect is separately defined to allow for cleaner composition.
    """

    analysis_path: str

    def output(self) -> DataTypes: ...

    def set_save_path(
        self,
        path: str | pathlib.Path,
        cache_path: str | pathlib.Path | None = None,
    ) -> None: ...



__doc__ = """
Connectivity module for localized instantaneous analysis
"""
__all__ = ["InstantaneousConnectivity"]


import itertools
import logging
import os
from dataclasses import dataclass

import matplotlib.animation as manimation
import matplotlib.pyplot as plt
import numpy as np
import pyinform
import pyinform.transferentropy as pyte
from tqdm import tqdm

from miv.core.datatype import Spikestamps
from miv.core.operator.operator import OperatorMixin
from miv.core.operator.policy import StrictMPIRunner


@dataclass
class InstantaneousConnectivity(OperatorMixin):
    """ """

    channels: list[int] | None = None
    bin_size: float = 0.001
    tag: str = "instantaneous connectivity rendering"
    progress_bar: bool = False

    te_history: int = 21

    fps: int = 25

    def __post_init__(self):
        super().__init__()
        self.runner = StrictMPIRunner()
        if self.runner.is_root:
            os.makedirs(self.analysis_path, exist_ok=True)
        self.runner.comm.Barrier()

    def __call__(self, spikestamps: Spikestamps) -> np.ndarray:
        """__call__.

        Parameters
        ----------
        spikestamps : Spikestamps
        """
        # comm = self.runner.comm
        rank = self.runner.get_rank()
        size = self.runner.get_size()

        binned_spiketrain = spikestamps.binning(bin_size=self.bin_size)
        n_timesteps = binned_spiketrain.shape[binned_spiketrain._SIGNALAXIS]

        # Channel Selection
        if self.channels is None:
            n_nodes = binned_spiketrain.number_of_channels
            channels = tuple(range(n_nodes))
        else:
            n_nodes = len(self.channels)
            channels = tuple(self.channels)

        # Define task
        pairs = [
            (i, j) for i, j in itertools.product(channels, channels) if i != j and i < j
        ]
        ntasks = len(pairs)

        # Split tasks
        if size > ntasks:
            logging.warning(
                f"Too many ranks. {size - ntasks} number of ranks will be idle."
            )
        tasks = np.array_split(pairs, size)[rank]

        # Run tasks
        disable_tqdm = not self.progress_bar or not self.runner.is_root
        for i, j in tqdm(tasks, disable=disable_tqdm):
            # compute i -> j
            ij_te = self._compute_local_directionality(
                binned_spiketrain[i],
                binned_spiketrain[j],
                self.te_history,
            )[0]
            # compute j -> i
            ji_te = self._compute_local_directionality(
                binned_spiketrain[j],
                binned_spiketrain[i],
                self.te_history,
            )[0]

            FFMpegWriter = manimation.writers["ffmpeg"]
            metadata = dict(
                title="Movie Test", artist="Matplotlib", comment="Movie support!"
            )
            writer = FFMpegWriter(fps=self.fps, metadata=metadata)
            video_name = os.path.join(
                self.analysis_path, f"local_te_channel_{i}_{j}.mp4"
            )

            # plot 3x1 figure: imshow of binnec spiketrain for i and j, ij_te, ji_te
            fig = plt.figure(figsize=(16, 5))
            sidx = 0
            dframe = int(1 / self.bin_size / self.fps)
            window = int(5.0 / self.bin_size)
            pbar = tqdm(total=n_timesteps, disable=disable_tqdm)
            with writer.saving(fig, video_name, dpi=200):
                while sidx + window < n_timesteps:
                    stime = binned_spiketrain.get_start_time() + sidx * self.bin_size
                    etime = stime + window * self.bin_size
                    time = binned_spiketrain.timestamps[sidx : sidx + window]

                    fig.clf()
                    ax1 = fig.add_subplot(311)
                    _lineoffsets = [0, 1]
                    _linelengths = [1, 1]
                    ax1.eventplot(
                        spikestamps.select([i, j], keepdims=False).get_view(
                            stime, etime
                        ),
                        lineoffsets=_lineoffsets,
                        linelengths=_linelengths,
                    )
                    ax1.set_yticklabels(["i", "j"])
                    ax1.set_xlim([stime, etime])

                    ax2 = fig.add_subplot(312)
                    ax2.plot(time, ij_te[sidx : sidx + window])
                    ax2.set_ylabel("i -> j")
                    ax2.set_ylim([ij_te.min(), ij_te.max()])

                    ax3 = fig.add_subplot(313)
                    ax3.plot(time, ji_te[sidx : sidx + window])
                    ax3.set_ylabel("j -> i")
                    ax3.set_ylim([ji_te.min(), ji_te.max()])

                    ax2.sharex(ax1)
                    ax3.sharex(ax1)

                    # Make it tight layout
                    fig.tight_layout()
                    writer.grab_frame()

                    sidx += dframe
                    pbar.update(dframe)
            plt.close(plt.gcf())
            pbar.close()

    def _compute_local_directionality(self, source, target, te_history):
        te = pyte.transfer_entropy(source, target, te_history, local=True)
        normalizer = pyinform.entropyrate.entropy_rate(
            target, k=te_history, local=False
        )
        if np.isclose(normalizer, 0.0):
            return np.zeros_like(te)
        else:
            return te / normalizer



__doc__ = """
Avalanche Analysis
"""

__all__ = ["AvalancheDetection", "AvalancheAnalysis"]


import logging
import os
from collections import defaultdict
from dataclasses import dataclass

import matplotlib.pyplot as plt
import numpy as np
import quantities as pq
from scipy.optimize import curve_fit
from tqdm import tqdm

from miv.core.datatype import Spikestamps
from miv.core.operator.operator import OperatorMixin
from miv.core.operator.wrapper import cache_call


@dataclass
class AvalancheDetection(OperatorMixin):
    r"""
    An operator for finding avalanches in spike trains.
    It returns the size, duration, and avalanches themselves, defined as all the times each neuron spiked in a given avalanche.

    Parameters
    ----------
    bin_size : float
        The size of the bins in seconds. Default is 0.004 seconds.
    threshold_percentage : float
        The threshold to detect avalanche.
        Default is 2.0, which means the threshold is the mean of the population firing rate divided by 2.
        Higher the value, less avalanches will be detected, and the avalanche size will be smaller.
    time_difference : float
        Minimum time difference between avalanches in seconds.
        Default is equal to the bin_size.
    allow_multiple_spike_per_bin : bool
        If True, allow multiple spikes per bin. Default is False.
        This influence the threshold of the avalanche size.
    minimum_bins_in_avalanche : int
        The minimum number of bins in an avalanche. Default is 10.
    min_interburst_interval_bound : float
        The minimum interburst interval in seconds. If burst interval is shorter, two burst will be coalaced. Default is 0.1.

    Returns
    -------
    starts : np.ndarray
        The start timestamp of each avalanche.
    ends : np.ndarray
        The end timestamp of each avalanche.
    bincount: Signal
        The binned spike count.

        .. math::
            \sum(thisShape[1:] / thisShape[:-1]) / Avalanche.duration
    """

    bin_size: float = 0.002  # in seconds

    # Miscellaneous configurations
    threshold_percentage: float = 2.0
    time_difference: float | None = None
    allow_multiple_spike_per_bin: bool = False
    minimum_bins_in_avalanche: int = 10

    min_interburst_interval_bound: float = 0.1  # sec
    pre_burst_extension: float = 0.0
    post_burst_extension: float = 0.0

    tag: str = "avalanche detection"

    @cache_call
    def __call__(self, spikestamps: Spikestamps):
        bincount = spikestamps.binning(
            bin_size=self.bin_size, return_count=self.allow_multiple_spike_per_bin
        )
        population_firing = bincount.data.sum(
            axis=bincount._CHANNELAXIS
        )  # Spike count accross channel per bin
        # threshold = (
        #    population_firing[np.nonzero(population_firing)].mean()
        #    / self.threshold_percentage
        # )
        non_zero_firing = population_firing[np.nonzero(population_firing)]
        mad = np.median(np.abs(non_zero_firing - np.median(non_zero_firing)))
        threshold = 1.4826 * mad * 4.0
        # TODO: try to reuse the code from miv.statistics.burst.burst
        events = (population_firing > threshold).astype(np.bool_)

        # pad and find where avalanches start or end based on where 0s change to
        # 1s and vice versa
        concat_events = np.concatenate([[False], events, [False]])
        diff_events = concat_events[1:].astype(np.int_) - concat_events[:-1].astype(
            np.int_
        )
        starts = np.where(diff_events == 1)[0]
        ends = np.where(diff_events == -1)[0]

        if len(starts) == 0 or len(ends) == 0:
            return np.array([]), np.array([]), bincount

        # remove avalanches that are too short
        if self.minimum_bins_in_avalanche > 1:
            inds = np.where(ends - starts >= self.minimum_bins_in_avalanche)[0]
            starts = starts[inds]
            ends = ends[inds]

        # TODO: duplicated lines
        if len(starts) == 0 or len(ends) == 0:
            return np.array([]), np.array([]), bincount

        if not np.isclose(self.bin_size, self.time_difference):
            starts2 = starts[1:]
            ends2 = ends[:-1]
            avalanche_interval = (starts2 - ends2) * self.bin_size
            inds = np.arange(len(avalanche_interval))
            inds = inds[avalanche_interval > self.time_difference]

            starts = starts[np.concatenate([[0], inds + 1])]
            ends = ends[np.concatenate([inds, [len(ends) - 1]])]

        # Include residual windows
        starts = starts - int(self.pre_burst_extension / self.bin_size)
        ends = ends + int(self.post_burst_extension / self.bin_size)

        # Coalace overlapped intervals
        while True:
            coalace_index = np.where(
                starts[1:]
                <= ends[:-1] + self.min_interburst_interval_bound / self.bin_size
            )[0]
            if coalace_index.size == 0:
                break
            starts = np.delete(starts, coalace_index + 1)
            ends = np.delete(ends, coalace_index)

        return starts, ends, bincount

    def __post_init__(self):
        super().__init__()
        if isinstance(self.bin_size, pq.Quantity):
            self.bin_size = self.bin_size.rescale("s").magnitude
        if isinstance(self.time_difference, pq.Quantity):
            self.time_difference = self.time_difference.rescale("s").magnitude
        if self.time_difference is None:
            self.time_difference = self.bin_size

    def plot_avalanche_on_raster(self, outputs, inputs, show=False, save_path=None):
        """Plot firing rate histogram"""
        if show:
            logging.warning(
                "Plotting avalanche on raster is not supported to be shown interactively."
            )
        spikestamps = self.receive()[0]
        starts, ends, bincount = outputs
        starts_time = starts * self.bin_size + bincount.timestamps[0]
        ends_time = ends * self.bin_size + bincount.timestamps[0]

        fig, ax = plt.subplots(figsize=(16, 6))
        ax.eventplot(spikestamps)

        for start, end in zip(starts_time, ends_time, strict=False):
            ax.axvspan(start, end, facecolor="r", alpha=0.5)

        ax.set_title("Avalanche Marks on Raster")
        ax.set_xlabel("Time (s)")
        ax.set_ylabel("Channel")

        # Save
        interval = 10  # sec TODO
        left, right = (
            spikestamps.get_first_spikestamp(),
            spikestamps.get_last_spikestamp(),
        )
        left -= interval * 0.1
        right += interval * 0.1
        for i in range(int(np.ceil((right - left) / interval))):
            ax.set_xlim(left + i * interval, left + (i + 1) * interval)
            if save_path is not None:
                plt.savefig(
                    os.path.join(save_path, f"raster_avalanche_overlay_{i}.png")
                )

            if i > 10:  # For early finish
                break
        plt.close(fig)


@dataclass
class AvalancheAnalysis(OperatorMixin):
    """
    Avalanche and Criticality Analysis

    .. example::

        >>> from miv.statistics.criticality import AvalancheDetection, AvalancheAnalysis

        >>> # Avalanche detection
        >>> avalanche_detection = AvalancheDetection()
        >>> avalanche_analysis = AvalancheAnalysis()
        >>> avalanche_detection >> avalanche_analysis

    Parameters
    ----------
    tag: str
        The tag for the operator.
    """

    tag: str = "avalanche criticality analysis"

    progress_bar: bool = False

    @cache_call
    def __call__(self, inputs):
        starts, ends, bincount = inputs
        bin_size = 1.0 / bincount.rate
        durations = (ends - starts) * bin_size

        # Compute size and branching ratio
        size = np.zeros(starts.size, dtype=np.int_)
        branching_ratio = np.zeros(starts.size, dtype=np.float64)
        avalanches = []
        for idx, (s, e) in tqdm(
            enumerate(zip(starts, ends, strict=False)),
            total=len(starts),
            desc="Avalanche analysis",
            disable=not self.progress_bar,
        ):
            if s == e:
                raise ValueError("Avalanche size is 0")
            avalanche = bincount.data[s:e, :]
            avalanches.append(avalanche)
            size[idx] = np.count_nonzero(avalanche.sum(axis=0))
            shape = avalanche.sum(axis=1)
            if np.isclose(size[idx], 0):
                branching_ratio[idx] = 0.0
            else:
                mask = shape[:-1] > 1e-5
                branching_ratio[idx] = (
                    np.sum(shape[1:][mask] / shape[:-1][mask]) / size[idx]
                )

        return durations, size, branching_ratio, avalanches, bin_size

    def __post_init__(self):
        super().__init__()

    def plot_power_law_fitting(self, outputs, inputs, show=False, save_path=None):
        durations, size, branching_ratio, avalanches, bin_size = outputs

        def power(x, a, c):
            return c * (x**a)

        def neg_power(x, a, c):
            return c * (x ** (-a))

        nbins = 100

        fig, axes = plt.subplots(1, 3, figsize=(18, 6))
        hist, bins = np.histogram(size, bins=nbins)
        logbins = np.logspace(np.log10(bins[0]), np.log10(bins[-1]), len(bins))
        axes[0].hist(size, bins=logbins, histtype="step", label="data")
        axes[0].set_xscale("log")
        axes[0].set_yscale("log")
        axes[0].set_xlabel("size (# channels)")
        axes[0].set_ylabel("Event Frequency")
        hist, bins = np.histogram(size, bins=logbins)
        if hist[hist > 1].size > 0:
            try:
                popt, pcov = curve_fit(neg_power, bins[:-1][hist > 1], hist[hist > 1])
                tau = popt[0]
                axes[0].plot(
                    logbins, neg_power(logbins, *popt), label=f"fit {tau=:.2f}"
                )
            except RuntimeError:
                tau = 0
                logging.warning("Power-fit failed. No fitted line will be plotted.")
            except TypeError:
                tau = 0
                logging.warning("Power-fit failed. No fitted line will be plotted.")
            self.tau = tau
        else:
            tau = 0.0
            self.tau = 0.0
        axes[0].legend()
        axes[0].set_ylim(bottom=5e-1)

        hist, bins = np.histogram(durations, bins=nbins)
        logbins = np.logspace(np.log10(bins[0]), np.log10(bins[-1]), len(bins))
        axes[1].hist(durations, bins=logbins, histtype="step", label="data")
        axes[1].set_xscale("log")
        axes[1].set_yscale("log")
        axes[1].set_xlabel("duration (s)")
        axes[1].set_ylabel("Event Frequency")
        hist, bins = np.histogram(durations, bins=logbins)
        try:
            if (hist > 1).sum() > 0:
                popt, pcov = curve_fit(neg_power, bins[:-1][hist > 1], hist[hist > 1])
                alpha = popt[0]
                axes[1].plot(
                    logbins, neg_power(logbins, *popt), label=f"fit {alpha=:.2f}"
                )
            else:
                alpha = 0
        except RuntimeError:
            alpha = 0
            logging.warning("Power-fit failed. No fitted line will be plotted.")
        except TypeError:
            alpha = 0
            logging.warning("Power-fit failed. No fitted line will be plotted.")
        self.alpha = alpha
        axes[1].legend()
        axes[1].set_ylim(bottom=5e-1)

        def width(p, w):
            return 10 ** (np.log10(p) + w / 2.0) - 10 ** (np.log10(p) - w / 2.0)

        values = []
        avearges = []
        for value in np.unique(durations):
            positions = np.array([value])
            axes[2].boxplot(
                size[durations == value],
                positions=positions,
                widths=width(positions, 0.1),
                showfliers=False,
            )
            values.append(value)
            avearges.append(np.mean(size[durations == value]))
        axes[2].set_xscale("log")
        axes[2].set_yscale("log")
        axes[2].set_xlabel("duration (s)")
        axes[2].set_ylabel("Average size (# channels)")

        self.svz = None
        try:
            popt, pcov = curve_fit(power, values, avearges)
            self.svz = popt[0]
            axes[2].plot(
                logbins, power(logbins, *popt), label=f"fit 1/svz={popt[0]:.2f}"
            )
        except (RuntimeError, TypeError, ValueError):
            logging.warning("Power-fit failed. No fitted line will be plotted.")
        axes[2].legend()
        axes[2].set_title(f"({(alpha-1)/(tau-1)=:.2f})")
        self.svz_estim_ratio = (alpha - 1) / (tau - 1)

        if save_path is not None:
            plt.savefig(os.path.join(save_path, "avalanche_power_fitting.png"))
        plt.close(fig)

        _, _, _, avalanches, bin_size = outputs

        shapes = defaultdict(list)
        for avalanche in avalanches:
            count = avalanche.shape[0]
            if count <= 1:
                continue
            shapes[count].append(avalanche.sum(axis=1))

        fig, axes = plt.subplots(1, 2, figsize=(12, 6))
        for idx, count in enumerate(shapes.keys()):
            shapes[count] = np.array(shapes[count])
            time = np.arange(count) * bin_size
            T = count * bin_size
            mean = np.array(shapes[count]).mean(axis=0)
            err = np.array(shapes[count]).std(axis=0)
            axes[0].errorbar(
                time, mean, yerr=err, label=f"duration {bin_size * count * 1000:.2f}ms"
            )
            if self.svz is not None:
                axes[1].plot(
                    time / T,
                    mean / (T ** (self.svz - 1)),
                    label=f"duration {bin_size * count * 1000:.2f}ms",
                )

        axes[0].set_xlabel("Time in Avalanche (s)")
        axes[0].set_ylabel("Average Number of Firing s(t,T)")
        axes[0].set_title("Raw Shapes")
        # axes[0].legend()
        if self.svz is not None:
            axes[1].set_xlabel("Time in Avalanche Duration (s / T)")
            axes[1].set_ylabel("Average Number of Firing s(t,T) / T^1/svz-1 ")
            axes[1].set_title(f"Normalized/Collapsed Avalanche Shape {self.svz:.2f}")
            # axes[1].legend()
        else:
            axes[1].set_title("Normalized/Collapsed Avalanche Shape (No fit)")

        if save_path is not None:
            plt.savefig(os.path.join(save_path, "avalanche_shape_collapse.png"))
        if show:
            plt.show()
        plt.close(fig)

    def plot_branching_ratio_histogram(
        self, outputs, inputs, show=False, save_path=None
    ):
        _, _, branching_ratio, _, _ = outputs

        nbins = 100

        fig, ax = plt.subplots(1, 1)
        # hist, bins = np.histogram(size, bins=nbins)
        # logbins = np.logspace(np.log10(bins[0]), np.log10(bins[-1]), len(bins))
        self.mean_branching_ratio = branching_ratio[np.nonzero(branching_ratio)].mean()
        ax.hist(
            branching_ratio[np.nonzero(branching_ratio)],
            bins=nbins,
            histtype="step",
            label="data",
        )
        ax.set_xlabel("branching ratio")
        ax.set_ylabel("Event Frequency")
        ax.set_title("branching ratio")
        if save_path is not None:
            plt.savefig(os.path.join(save_path, "branching_ratio.png"))
        if show:
            plt.show()
        plt.close(fig)



__doc__ = """
Topological sorting of the graph.
"""

__all__ = ["topological_sort"]

from ..operator.protocol import _Node


def _get_upstream_topology(
    node: _Node, upstream_nodelist: list[_Node] | None = None
) -> list[_Node]:
    """
    Get the upstream topology of a node.

    Args:
        node: The node to get the upstream topology for
        upstream_nodelist: Optional list of already visited nodes

    Returns:
        List of nodes in the upstream topology
    """
    if upstream_nodelist is None:
        upstream_nodelist = []

    # Check if node is cachable
    cached_flag = False
    try:
        cached_flag = node.cacher.check_cached()
    except (AttributeError, FileNotFoundError):
        """
        For any reason when cached result could not be retrieved.

        AttributeError: Occurs when cacher is not defined
        FileNotFoundError: Occurs when cache_dir is not set or cache files doesn't exist
        """
        pass

    if not cached_flag:  # Run all upstream nodes
        for upstream_node in node.iterate_upstream():
            if upstream_node in upstream_nodelist:
                continue
            _get_upstream_topology(upstream_node, upstream_nodelist)
    upstream_nodelist.append(node)
    return upstream_nodelist


def topological_sort(node: _Node) -> list[_Node]:
    """
    Topological sort of the graph.
    Returns the list of operations in order to execute the given node.

    Args:
        node: The node to get the topological sort for

    Returns:
        List of nodes in topological order

    Raises:
        RuntimeError: If there is a loop in the graph.
    """
    upstream = _get_upstream_topology(node)

    # Track indices of nodes in the sorted list
    key = []
    pos = []
    ind = 0
    tsort: list[_Node] = []

    while len(upstream) > 0:
        key.append(upstream[-1])
        pos.append(ind)
        tsort.append(upstream[-1])
        ind += 1
        upstream.pop()

    for source in tsort:
        for up in source.iterate_upstream():
            if up not in key:
                continue
            before = pos[key.index(source)]
            after = pos[key.index(up)]

            # If parent vertex does not appear first
            if before > after:
                raise RuntimeError(
                    f"Found loop in operation stream: node {source} is already in the upstream : {up}."
                )

    tsort.reverse()
    return tsort



__doc__ = "Downsample Operator for Signals"
__all__ = ["Downsample"]

from dataclasses import dataclass

import numpy as np
import scipy.signal as sps

from miv.core.datatype import Signal
from miv.core.operator_generator.operator import GeneratorOperatorMixin
from miv.core.operator_generator.wrapper import cache_generator_call


@dataclass
class Downsample(GeneratorOperatorMixin):
    """Downsample the signal to the target sampling rate, wrapped in operator mixin.

    Parameters
    ----------
    target_rate : float
        Target sampling rate (Hz) for downsampling.
    """

    target_rate: float = 1000.0  # Hz
    tag: str = "downsample operator"

    @cache_generator_call
    def __call__(self, signal: Signal) -> Signal:
        """Downsample the signal to the target sampling rate.

        Parameters
        ----------
        signal : SignalType
            signal
        """
        # Calculate downsampling factor
        downsample_factor = int(signal.rate / self.target_rate)
        if downsample_factor < 1:
            raise ValueError(
                "Invalid downsample factor. Check target and original rates."
            )

        # Downsample each channel
        data = signal.data
        downsampled_data = sps.resample(
            data, num=data.shape[0] // downsample_factor, axis=0
        )

        # Adjust timestamps for the new sampling rate
        downsampled_timestamps = np.linspace(
            signal.timestamps[0], signal.timestamps[-1], downsampled_data.shape[0]
        )

        return Signal(
            data=downsampled_data,
            timestamps=downsampled_timestamps,
            rate=self.target_rate,
        )

    def __post_init__(self):
        super().__init__()
        # self.cacher.policy = "OFF"



from __future__ import annotations

__doc__ = """
"""
__all__ = [
    "DefaultLoggerMixin",
]

from typing import TYPE_CHECKING, Any

import logging

if TYPE_CHECKING:
    pass


class DefaultLoggerMixin:
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        try:
            from mpi4py import MPI

            comm = MPI.COMM_WORLD
            if comm.Get_size() > 1:
                tag = f"rank[{comm.Get_rank()}]-{self.__class__.__name__}"
            else:
                tag = self.__class__.__name__
        except ImportError:
            tag = self.__class__.__name__
        self._logger = logging.getLogger(tag)

    @property
    def logger(self) -> logging.Logger:
        return self._logger

    # TODO: Redirect I/O stream to txt



import matplotlib.pyplot as plt
import numpy as np

from miv.core.datatype import Spikestamps
from miv.statistics import burst
from miv.typing import SpikestampsType


def plot_spiketrain_raster(
    spikestamps: Spikestamps, t_start: float, t_stop: float
):  # pragma: no cover
    """
    Plot spike train in raster

    Parameters
    ----------
    spikestamps : Spikestamps
        Spike stamps
    t_start : float
        Start time
    t_stop : float
        Stop time
    """
    fig, ax = plt.subplots(figsize=(16, 6))
    spikes = spikestamps.get_view(t_start, t_stop)
    print(
        f"Plotting raster plot from {t_start} to {t_stop}: {sum(spikes.get_count())} spikes"
    )
    ax.eventplot(spikes)
    ax.set_xlabel("Time (s)")
    ax.set_ylabel("Channel")
    ax.set_title(f"Raster plot (from {t_start} to {t_stop})")
    ax.set_xlim(t_start, t_stop)
    return fig, ax


def plot_burst(
    spiketrains: SpikestampsType, min_isi: float, min_len: float
):  # pragma: no cover
    """
    Plots burst events across electrodes  to characterize bursting phenomenon on a singl channel

    Parameters
    ----------
    spikes : SpikestampsType
           Single spike-stamps
    min_isi : float
       Minimum Interspike Interval (in seconds) to be considered as bursting [standard = 0.1]
    min_len : float
       Minimum number of simultaneous spikes to be considered as bursting [standard = 10]

    Returns
    -------
    figure, axes
       matplot figure with bursts plotted for all electrodes
    """
    fig, ax = plt.subplots()
    start_time = []
    burst_duration = []
    burst_len = []
    burst_rate = []
    for i in np.arange(len(spiketrains)):
        start_time, burst_duration, burst_len, burst_rate = burst(
            spiketrains, i, min_isi, min_len
        )
        a = np.column_stack((start_time, burst_duration))
        ax.broken_barh(a, (i, 0.5), facecolors="tab:orange")

    times = np.concatenate([train for train in spiketrains])
    channel = np.concatenate(
        [np.full_like(train, i, dtype=np.float_) for i, train in enumerate(spiketrains)]
    )
    ax.scatter(times, channel, s=2, marker="o", color="blue")
    ax.set_xlabel("Time (s)")
    ax.set_ylabel("Electrode")

    return fig, ax



__doc__ = """Multi-channel signal plotting for MEA channels"""
__all__ = ["MultiChannelSignalVisualization"]


import inspect
import os
from dataclasses import dataclass

import matplotlib.animation as manimation
import numpy as np
from matplotlib import pyplot as plt
from tqdm import tqdm

from miv.core.datatype import Spikestamps
from miv.core.operator.operator import OperatorMixin
from miv.mea.protocol import MEAGeometryProtocol


@dataclass
class MultiChannelSignalVisualization(OperatorMixin):
    average_interval: int = 50  # frames to skip. TODO: refactor

    tag: str = "signal render"

    progress_bar: bool = False

    fps: int = 25
    dpi: int = 200

    def __post_init__(self):  # pragma: no cover
        super().__init__()

    def __call__(
        self, signals: Spikestamps, mea: MEAGeometryProtocol
    ):  # pragma: no cover
        if not inspect.isgenerator(signals):
            signals = [signals]

        for vidx, signal in enumerate(signals):
            probe_times = signal.timestamps[:: self.average_interval]
            n = (signal.data.shape[0] // self.average_interval) * self.average_interval
            xs = np.average(
                signal.data[:n, :].reshape(
                    -1, self.average_interval, signal.number_of_channels
                ),
                axis=1,
            )
            med = np.median(np.abs(xs))
            xmax = 15.0 * med
            xmin = -15.0 * med

            # Output Images
            FFMpegWriter = manimation.writers["ffmpeg"]
            metadata = dict(
                title="Movie Test", artist="Matplotlib", comment="Movie support!"
            )
            writer = FFMpegWriter(fps=self.fps, metadata=metadata)

            os.makedirs(self.analysis_path, exist_ok=True)
            video_name = os.path.join(self.analysis_path, f"output_{vidx}.mp4")

            fig = plt.figure(figsize=(8, 6))
            with writer.saving(fig, video_name, dpi=self.dpi):
                for timestep in tqdm(
                    range(xs.shape[0]),
                    desc="Rendering",
                    total=xs.shape[0],
                    disable=not self.progress_bar,
                ):
                    time = probe_times[timestep]
                    X, Y, Z = mea.map_data(xs[timestep, :])

                    fig.clf()
                    ax = fig.add_subplot(111)
                    # X, Y, Z = interp_2d(Z)
                    ax.plot(mea.coordinates[:, 0], mea.coordinates[:, 1], "k.", ms=1)
                    pcm = ax.pcolormesh(
                        X,
                        Y,
                        Z,
                        cmap="MGBlueOrange",
                        vmin=xmin,
                        vmax=xmax,
                        shading="gouraud",
                    )
                    cbar = fig.colorbar(pcm, ax=ax)
                    cbar.ax.set_ylabel(
                        f"signal averaged over {self.average_interval / signal.rate:.02f} sec",
                        rotation=270,
                    )

                    ax.set_aspect("equal")
                    ax.set_xlabel("channels x-axis (µm)")
                    ax.set_ylabel("channels y-axis (µm)")
                    ax.set_title(f"Signal ({time:.05f} sec)")
                    ax.invert_yaxis()

                    writer.grab_frame()
            plt.close(plt.gcf())



__all__ = ["PythonDataType", "NumpyDType", "GeneratorType"]

from typing import TypeAlias, Any
from collections.abc import Generator, Iterator

import numpy as np

from miv.core.operator.operator import DataNodeMixin
from miv.core.operator.chainable import BaseChainingMixin

PurePythonTypes: TypeAlias = int | float | str | bool | list | tuple | dict


class ValuesMixin(DataNodeMixin, BaseChainingMixin):
    """
    This mixin is used to convert pure/numpy data type to be a valid input/output of a node.
    """

    def __init__(
        self, data: np.ndarray | PurePythonTypes, *args: Any, **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.data = data


class PythonDataType(ValuesMixin):
    """
    Python data types are: int, float, str, bool, list, tuple, dict
    This type also allows None.
    """

    @staticmethod
    def is_valid(data: Any) -> bool:
        return data is None or isinstance(
            data, int | float | str | bool | list | tuple | dict
        )


class NumpyDType(ValuesMixin):
    """
    Numpy data types.
    """

    @staticmethod
    def is_valid(data: Any) -> bool:
        return isinstance(data, np.ndarray)


class GeneratorType(BaseChainingMixin):
    def __init__(self, iterator: Iterator, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.iterator = iterator

    def output(self) -> Generator:
        yield from self.iterator

    def run(self, **kwargs: Any) -> Generator:
        yield from self.output()

    @staticmethod
    def is_valid(data: Any) -> bool:
        import inspect

        return inspect.isgenerator(data)



__all__ = ["ImportSignal"]

from collections.abc import Generator


from miv.core.datatype import Signal
from miv.core.operator.operator import DataLoaderMixin
from miv.io import file as miv_file


class ImportSignal(DataLoaderMixin):
    def __init__(
        self,
        data_path: str,
        tag: str = "import signal",
    ) -> None:
        self.data_path: str = data_path
        super().__init__()
        self.tag: str = f"{tag}"

    def load(self) -> Generator[Signal]:
        data, container = miv_file.read(self.data_path)
        num_container = data["_NUMBER_OF_CONTAINERS_"]
        self.logger.info(f"Loading: {num_container=}")

        for i in range(num_container):
            miv_file.unpack(container, data, i)
            signal = Signal(
                data=container["Ephys/Data"],
                timestamps=container["Ephys/Timestamps"],
                rate=container["Ephys/Rate"],
            )
            self.logger.info(
                f"{i}-container | {signal.data.shape=}, {signal.timestamps.shape=}, {signal.rate=}"
            )
            yield signal



__doc__ = """

Module (MiV-Simulator)
######################

.. autoclass:: Data
   :members:

"""
__all__ = ["Data"]

from typing import (
    Any,
)
from collections.abc import Generator

import os

import h5py
import numpy as np

from miv.core.datatype.signal import Signal
from miv.core.operator.operator import DataLoaderMixin


class Data(DataLoaderMixin):
    """Single result from miv-simulator.

    Parameters
    ----------
    data_path : str

    """

    tag = "Simulation data loader"

    def __init__(self, data_path: str, *args: Any, **kwargs: Any) -> None:
        self.data_path = data_path
        super().__init__(*args, **kwargs)

        self._lfp_key = "Local Field Potential"  # TODO: refactor
        self._load_every = 60  # sec. Parse every 60 sec.

    def load(self) -> Generator[Signal]:
        yield from self.load_lfp_recordings()

    def load_lfp_recordings(
        self, indices: list[int] | None = None
    ) -> Generator[Signal]:
        infile = h5py.File(self.data_path)

        if indices is not None:
            keys = [
                key
                for key in infile.keys()
                if self._lfp_key in key and int(key.split(" ")[-1]) in indices
            ]
        else:
            keys = [key for key in infile.keys() if self._lfp_key in key]

        # Check if t matches
        t0: np.ndarray = np.array([])
        for _, namespace_id in enumerate(keys):
            if t0.size == 0:
                t0 = np.asarray(infile[namespace_id]["t"])
                continue
            t = np.asarray(infile[namespace_id]["t"])
            if not np.allclose(t0, t):
                raise ValueError(
                    "Recorded time for electrodes does not match."
                    "Check if the sampling rates for each electrode are the same."
                )

        sampling_rate = float(
            1000.0 / np.median(np.diff(t0))
        )  # FIXME: Try to infer from environment configuration instead
        length = int(self._load_every * sampling_rate)
        findex = len(t0)
        sindex, eindex = 0, min(length, findex)
        while eindex <= findex:
            signals = []
            for _, namespace_id in enumerate(keys):
                v = np.asarray(infile[namespace_id]["v"])
                signals.append(v[sindex:eindex])
            yield Signal(
                data=np.asarray(signals).T,
                timestamps=t0[sindex:eindex],
                rate=sampling_rate,
            )
            sindex += length
            eindex = min(eindex + length, len(t0))

    def check_path_validity(self) -> bool:
        """
        Check if necessary files exist in the directory.

        Returns
        -------
        bool
            Return true if all necessary files exist in the directory.
        """
        return os.path.exists(self.data_path)



from __future__ import annotations

__doc__ = """
"""
__all__ = [
    "_CacherProtocol",
    "BaseCacher",
    "DataclassCacher",
    "FunctionalCacher",
]

from typing import TYPE_CHECKING, Any, Literal, Protocol
from collections.abc import Callable, Generator

from collections import OrderedDict
import dataclasses
import glob
import json
import os
import pathlib
import pickle as pkl
import shutil

import numpy as np

from miv.utils.formatter import TColors

if TYPE_CHECKING:
    from .protocol import _Cachable, _Node
    from miv.core.datatype import DataTypes

# ON: Always use cache if cache exist. Otherwise, run and save cache.
# OFF: Never use cache functions, always run. No cache save.
# MUST: must use cache. If no cache exist, raise error
# OVERWRITE: Always run and overwrite cache.
CACHE_POLICY = Literal["ON", "OFF", "MUST", "OVERWRITE"]


class _CacherProtocol(Protocol):
    policy: CACHE_POLICY
    cache_dir: str | pathlib.Path

    def __init__(self, parent: _Cachable) -> None: ...

    def load_cached(self, tag: str = "data") -> Generator[Any]:
        """Load the cached values."""
        ...

    def save_cache(self, values: Any, idx: int = 0, tag: str = "data") -> bool: ...

    def check_cached(self, tag: str = "data", *args: Any, **kwargs: Any) -> bool:
        """Check if the current configuration is the same as the cached one."""
        ...

    def save_config(self, tag: str = "data", *args: Any, **kwargs: Any) -> bool: ...


def when_policy_is(*allowed_policy: CACHE_POLICY) -> Callable:
    def decorator(
        func: Callable[[_CacherProtocol, Any, Any], bool | DataTypes],
    ) -> Callable:
        # @functools.wraps(func) # TODO: fix this
        def wrapper(
            self: _CacherProtocol, *args: Any, **kwargs: Any
        ) -> bool | DataTypes:
            if self.policy in allowed_policy:
                return func(self, *args, **kwargs)
            else:
                return False

        return wrapper

    return decorator


class BaseCacher:
    """
    Base class for cacher.
    """

    def __init__(self, parent: _Node) -> None:
        super().__init__()
        self.policy: CACHE_POLICY = "ON"
        self.parent = parent
        self.cache_dir: str | pathlib.Path = "results"

    def config_filename(self, tag: str = "data") -> str:
        return os.path.join(self.cache_dir, f"config_{tag}.json")

    def cache_filename(self, idx: int | str, tag: str = "data") -> str:
        index = idx if isinstance(idx, str) else f"{idx:04}"
        if getattr(self.parent.runner, "comm", None) is None:
            mpi_tag = f"{0:03d}"
        else:
            mpi_tag = f"{self.parent.runner.get_run_order():03d}"
        return os.path.join(self.cache_dir, f"cache_{tag}_rank{mpi_tag}_{index}.pkl")

    @when_policy_is("ON", "MUST", "OVERWRITE")
    def save_cache(self, values: Any, idx: int = 0, tag: str = "data") -> bool:
        os.makedirs(self.cache_dir, exist_ok=True)
        with open(self.cache_filename(idx, tag), "wb") as f:
            pkl.dump(values, f)
        return True

    def remove_cache(self) -> None:
        if os.path.exists(self.cache_dir):
            shutil.rmtree(self.cache_dir)

    def _load_configuration_from_cache(self, tag: str = "data") -> dict | str | None:
        path = self.config_filename(tag)
        if os.path.exists(path):
            with open(path) as f:
                return json.load(f)  # type: ignore[no-any-return]
        return None

    def log_cache_status(self, flag: bool) -> None:
        msg = f"Caching policy: {self.policy} - "
        if flag:
            msg += "Cache exist"
        else:
            msg += TColors.red + "No cache" + TColors.reset
        self.parent.logger.info(msg)


class DataclassCacher(BaseCacher):
    @when_policy_is("ON", "MUST", "OVERWRITE")
    def check_cached(self, tag: str = "data", *args: Any, **kwargs: Any) -> bool:
        if self.policy == "MUST":
            flag = True
        elif self.policy == "OVERWRITE":
            flag = False
        else:
            current_config = self._compile_configuration_as_dict()
            cached_config = self._load_configuration_from_cache(tag=tag)
            if cached_config is None:
                flag = False
            else:
                # Json equality
                flag = current_config == cached_config
        self.log_cache_status(flag)
        return flag

    def _compile_configuration_as_dict(self) -> dict[Any, Any]:
        config: OrderedDict = dataclasses.asdict(self.parent, dict_factory=OrderedDict)  # type: ignore
        for key in config.keys():
            if isinstance(config[key], np.ndarray):
                config[key] = config[key].tostring()
            elif hasattr(config[key], "to_json"):
                config[key] = config[key].to_json()
        return config

    @when_policy_is("ON", "MUST", "OVERWRITE")
    def save_config(self, tag: str = "data", *args: Any, **kwargs: Any) -> bool:
        config = self._compile_configuration_as_dict()
        os.makedirs(self.cache_dir, exist_ok=True)
        try:
            with open(self.config_filename(tag=tag), "w") as f:
                json.dump(config, f, indent=4)
        except (TypeError, OverflowError) as err:
            raise TypeError(
                "Some property of caching objects are not JSON serializable."
            ) from err
        return True

    def load_cached(self, tag: str = "data") -> Generator[DataTypes]:
        paths = glob.glob(self.cache_filename("*", tag=tag))
        paths.sort()
        for path in paths:
            with open(path, "rb") as f:
                self.parent.logger.info(f"Loading cache from: {path}")
                yield pkl.load(f)


class FunctionalCacher(BaseCacher):
    def _compile_parameters_as_dict(self, params: dict | None = None) -> dict:
        # Safe to assume params is a tuple, and all elements are hashable
        config: dict[str, Any] = OrderedDict()
        if params is None:
            return config
        for idx, arg in enumerate(params[0]):
            config[f"arg_{idx}"] = arg
        config.update(params[1])
        return config

    @when_policy_is("ON", "MUST", "OVERWRITE")
    def check_cached(self, params: dict | None = None, tag: str = "data") -> bool:
        if self.policy == "MUST":
            flag = True
        elif self.policy == "OVERWRITE":
            flag = False
        else:
            flag = True
            if params is not None:
                current_config = self._compile_parameters_as_dict(params)
                cached_config = self._load_configuration_from_cache(tag)

                if cached_config is None:
                    flag = False
                else:
                    # Json equality
                    flag = current_config == cached_config
            flag = flag and os.path.exists(self.cache_filename(0, tag=tag))
        self.log_cache_status(flag)
        return flag

    @when_policy_is("ON", "MUST", "OVERWRITE")
    def save_config(self, params: dict | None = None, tag: str = "data") -> bool:
        config = self._compile_parameters_as_dict(params)
        os.makedirs(self.cache_dir, exist_ok=True)
        try:
            with open(self.config_filename(tag), "w") as f:
                json.dump(config, f, indent=4)
        except (TypeError, OverflowError) as err:
            raise TypeError(
                "Some property of caching objects are not JSON serializable."
            ) from err
        return True

    def load_cached(self, tag: str = "data") -> Generator[DataTypes]:
        path = glob.glob(self.cache_filename(0, tag=tag))[0]
        with open(path, "rb") as f:
            self.parent.logger.info(f"Loading cache from: {path}")
            yield pkl.load(f)

    @when_policy_is("ON", "MUST", "OVERWRITE")
    def save_cache(self, values: Any, idx: int = 0, tag: str = "data") -> bool:
        os.makedirs(self.cache_dir, exist_ok=True)
        with open(self.cache_filename(0, tag=tag), "wb") as f:
            pkl.dump(values, f)
        return True



__all__ = ["PSTH", "PeristimulusActivity", "peri_stimulus_time", "PSTHOverlay"]


import itertools
import os
from dataclasses import dataclass

import matplotlib.pyplot as plt
import numpy as np

from miv.core.datatype import Signal, Spikestamps
from miv.core.operator.operator import OperatorMixin
from miv.core.operator.wrapper import cache_call
from miv.mea import mea_map


def peri_stimulus_time(spike_list):
    """
    Compute the peri-stimulus time of the given spike train.

    Examples
    --------

    How to draw Peri-stimulus time histogram (ISIH):

        >>> from miv.statistics import peri_stimulus_time
        >>> import matplotlib.pyplot as plt
        >>> pst = peri_stimulus_time(spikestamp)
        >>> plt.hist(pst)

    If one wants to get the bin-count, you can use `numpy.digitize` and `numpy.bincount`:

        >>> import numpy as np
        >>> max_time = spikestamps.max()
        >>> num_bins = 20
        >>> time_interval = np.linspace(0, max_time, num_bins)
        >>> digitized = np.digitize(pst, time_interval)
        >>> bin_counts = np.bincount(digitized)

    Parameters
    ----------
    spikes : SpikestampsType
        Single spike-stamps

    Returns
    -------
        interval: numpy.ndarray

    """

    peri_stimulus_times = np.sum(np.array(spike_list), 0)
    return peri_stimulus_times


@dataclass
class PeristimulusActivity(OperatorMixin):
    # Binning configuration
    # Default: 400ms domain, 4ms binsize
    mea: str = None
    interval: float = 0.4  # seconds
    tag: str = "peri-stimulus activity plot"

    stimulus_length: float = 0.010  # seconds. Skips for removing stimulus artifact

    def __post_init__(self):
        super().__init__()
        if isinstance(self.mea, str):
            self.mea_map = mea_map[self.mea]
        else:
            self.mea_map = mea_map["64_intanRHD"]

    @cache_call
    def __call__(self, events: Spikestamps, spikestamps: Spikestamps):
        # TODO: Change events datatype to be Event, not Spikestamps
        activity = [Spikestamps() for _ in range(spikestamps.number_of_channels)]
        for t_start in events[0]:
            t_end = t_start + self.interval
            view = spikestamps.get_view(
                t_start + self.stimulus_length, t_end + self.stimulus_length
            )
            for channel in range(view.number_of_channels):
                shifted_array = np.asarray(view[channel])
                if shifted_array.size > 0:  # FIXME: Optimize this
                    activity[channel].append(shifted_array - shifted_array[0])
                else:
                    activity[channel].append(shifted_array)
        return activity

    def plot_peristimulus_in_grid_map(
        self, activity, inputs, show=False, save_path=None
    ):
        mea_map = self.mea_map
        nrow, ncol = mea_map.shape
        fig, axes = plt.subplots(
            nrow, ncol, figsize=(ncol * 4, nrow * 4), sharex=True, sharey=True
        )
        for channel in range(len(activity)):
            p = activity[channel]
            if channel not in mea_map:
                continue
            w = np.where(mea_map == channel)
            r = w[0][0]
            c = w[1][0]
            axes[r][c].eventplot(p)
            axes[r][c].set_title(f"channel {channel}")
        # Bottom row
        for i in range(ncol):
            axes[-1, i].set_xlabel("time (s)")

        # Left row
        for i in range(nrow):
            axes[i, 0].set_ylabel("Stimulus Index")

        plt.suptitle("Peri-Stimulus spike activity for each channel")

        if show:
            plt.show()
        if save_path is not None:
            plt.savefig(os.path.join(save_path, "peristimulus_activity.png"))

        return axes


@dataclass
class PSTH(OperatorMixin):
    # Histogram Configuration
    # Default: 400ms domain, 4ms binsize
    mea: str = None
    binsize: float = 0.004  # seconds
    interval: float = 0.4  # seconds
    tag: str = "peri-stimulus time histogram"

    stimulus_length: float = 0.010  # seconds. Skips for removing stimulus artifact

    def __post_init__(self):
        super().__init__()
        if isinstance(self.mea, str):
            self.mea_map = mea_map[self.mea]
        else:
            self.mea_map = mea_map["64_intanRHD"]

    @cache_call
    def __call__(self, events: Spikestamps, spikestamps: Spikestamps):
        # TODO: Change events datatype to be Event, not Spikestamps
        n_time = int(np.ceil(self.interval / self.binsize))
        time_axis = np.linspace(0, self.interval, n_time) + self.stimulus_length
        psth = np.zeros((spikestamps.number_of_channels, n_time), dtype=np.float64)
        for t_start in events[0]:
            t_end = t_start + n_time * self.binsize
            bst = spikestamps.binning(
                self.binsize,
                t_start=t_start + self.stimulus_length,
                t_end=t_end + self.stimulus_length,
                return_count=False,
            )
            for channel in range(spikestamps.number_of_channels):
                psth[channel] += bst[channel][:n_time]
        psth /= len(events[0])
        psth /= self.binsize
        return Signal(data=psth.T, timestamps=time_axis, rate=1.0 / self.binsize)

    def plot_psth_in_grid_map(self, psth, inputs, show=False, save_path=None):
        mea_map = self.mea_map
        nrow, ncol = mea_map.shape
        fig, axes = plt.subplots(
            nrow, ncol, figsize=(ncol * 4, nrow * 4), sharex=True, sharey=True
        )
        for channel in range(psth.number_of_channels):
            p = psth[channel]
            time = psth.timestamps
            if channel not in mea_map:
                continue
            w = np.where(mea_map == channel)
            r = w[0][0]
            c = w[1][0]
            axes[r][c].plot(time, p)
            axes[r][c].set_title(f"channel {channel}")
        # Bottom row
        for i in range(ncol):
            axes[-1, i].set_xlabel("time (s)")

        # Left row
        for i in range(nrow):
            axes[i, 0].set_ylabel("mean (channels) spike rate per bin")

        plt.suptitle("PSTH: stimulating electrode")

        if show:
            plt.show()
        if save_path is not None:
            plt.savefig(os.path.join(save_path, "psth.png"))

        return axes


@dataclass
class PSTHOverlay(OperatorMixin):
    # TODO: Experimental
    # Histogram Configuration
    # Default: 400ms domain, 4ms binsize
    mea: str = None
    tag: str = "peri-stimulus time histogram overlay"

    stimulus_length: float = 0.010

    def __post_init__(self):
        super().__init__()
        if isinstance(self.mea, str):
            self.mea_map = mea_map[self.mea]
        else:
            self.mea_map = mea_map["64_intanRHD"]

    def __call__(self, *psths):
        return psths

    def plot_psth_in_grid_map(
        self,
        psths,
        inputs,
        show=False,
        save_path=None,
    ):
        mea_map = self.mea_map
        nrow, ncol = mea_map.shape
        fig, axes = plt.subplots(
            nrow, ncol, figsize=(ncol * 4, nrow * 4), sharex=True, sharey=True
        )
        for idx, psth in enumerate(psths):
            for channel in range(psth.number_of_channels):
                p = psth[channel]
                time = psth.timestamps
                if channel not in mea_map:
                    continue
                w = np.where(mea_map == channel)
                r = w[0][0]
                c = w[1][0]
                axes[r][c].plot(time, p, label=f"PSTH {idx}")
                axes[r][c].set_title(f"channel {channel}")
        # Bottom row
        for i in range(ncol):
            axes[-1, i].set_xlabel("time (s)")

        # Left row
        for i in range(nrow):
            axes[i, 0].set_ylabel("mean spike/bin")

        # Legend
        for i, j in itertools.product(range(nrow), range(ncol)):
            axes[i, j].legend(loc="best")

        plt.suptitle("PSTH: stimulating electrode")

        if show:
            plt.show()
        if save_path is not None:
            plt.savefig(os.path.join(save_path, "psth.png"))

        return axes

    def plot_psth_area_trend(
        self,
        psths,
        inputs,
        show=False,
        save_path=None,
    ):
        mea_map = self.mea_map
        nrow, ncol = mea_map.shape
        fig, axes = plt.subplots(
            nrow, ncol, figsize=(ncol * 4, nrow * 4), sharex=True, sharey=True
        )
        for channel in range(psths[0].number_of_channels):
            if channel not in mea_map:
                continue
            w = np.where(mea_map == channel)
            r = w[0][0]
            c = w[1][0]

            hist = []
            for idx, psth in enumerate(psths):
                p = psth[channel]
                time = psth.timestamps
                hist.append(np.trapz(p, time))
            axes[r][c].plot(hist)
            axes[r][c].set_title(f"channel {channel}")
        # Bottom row
        for i in range(ncol):
            axes[-1, i].set_xlabel("data points")

        # Left row
        for i in range(nrow):
            axes[i, 0].set_ylabel("Area under PSTH")

        plt.suptitle("PSTH Trend")

        if show:
            plt.show()
        if save_path is not None:
            plt.savefig(os.path.join(save_path, "psth_trend.png"))

        return axes
